/* 
 * pe_encoding.h
 * ProteinEncoder
 *
 * Created by Naoto Morikawa on 6/01/05.
 * Copyright 2005, 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-02 move PE_CODE to main.h
 *  2006-11-20 version 0.2 release
 *  2007-01-12 add my_settings.try_all_type (remove TRY_ALL_TYPE)
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#define ALL_ENC 0

#define MAX_LOCAL_AVE_TBL 20
#define NP_TBL_SIZE 6
#define NP_TBL_DISP 2

#define XX 0
#define YY 1
#define ZZ 2
#define WW 3
#define  INDX(p,q,r,s)    (p)+4*(q)+16*(r)+64*(s)

#define XYZW 0
#define XZYW 1
#define YXZW 2
#define YZXW 3
#define ZXYW 4
#define ZYXW 5

#define WXYZ 6
#define YWXZ 7
#define WYXZ 8
#define XWYZ 9
#define XYWZ 10
#define YXWZ 11

#define ZWXY 12
#define WXZY 13
#define XZWY 14
#define ZXWY 15
#define WZXY 16
#define XWZY 17

#define YZWX 18
#define ZYWX 19
#define ZWYX 20
#define WYZX 21
#define YWZX 22
#define WZYX 23

#define THIS_IS_NOT_CA_ATOM(atm_ct) \
	(my_settings.ca_only_flag == TRUE \
	&& my_protein.atom[(atm_ct)].atom_name != PE_CA)
#define THIS_IS_CA_ATOM(atm_ct) \
	(my_settings.ca_only_flag == TRUE \
	&& my_protein.atom[(atm_ct)].atom_name == PE_CA)
#define FLOAT_NEQ(fdata, dis) ((fdata) * (fdata) >  (dis) / 1000000.0)
#define NOT_REACHED_YET(cur_error) \
	((cur_error) >  \
	my_coor.ave_dis * (3.0 / my_settings.tiles_per_res \
			   + my_settings.dist_factor))
#define ALREADY_REACHED(cur_error) \
((cur_error) <=  \
	my_coor.ave_dis * (3.0 / my_settings.tiles_per_res \
			   + my_settings.dist_factor))
#define NOT_OVER_RUN(distance) \
	((distance) > my_coor.ave_dis * my_settings.dist_factor)
#define FAR_FROM_TARGET(cur_error) ((cur_error) > 1.0 * my_coor.ave_dis)

/*
typedef enum pe_code {
	C_UNKNOWN,
	C_UP,
	C_DOWN
} PE_CODE;
*/

typedef enum pe_direction {
	DIR_UNKOWN,
	DIR_FORWARD,
	DIR_BACKWARD
} PE_DIRECTION;

/**
 * struct pe_qtos_table - is a pair of quadruplet (x,y,z,w) and string "[xyz]"
 * @p  : the first literal of the type of a tile
 * @q  : the second literal of the type of a tile
 * @r  : the third literal of the type of a tile
 * @s  : the fourth literal of the type of a tile
 * @str: the string which describe the type of the tile
 */
struct pe_qtos_table {
	int p;
	int q;
	int r;
	int s;
	char str[5];
};

/**
 * struct pe_vector - is a three dimensional vector data
 * @x  : x-coordinate of the vector
 * @y  : y-coordinate of the vector
 * @z  : z-coordinate of the vector
 * @length: length of the vector, i.e., sqrt(x*x + y*y + z*z)
 */
struct pe_vector {
	float x;
	float y;
	float z;
	float length;
};

/**
 * struct pe_matrix - is a 3 by 3 matrix
 * @aij  : the (i,j) element
 *
 * a11 a12 a13
 * a21 a22 a23
 * a31 a32 a33
 */
struct pe_matrix {
	float a11;
	float a12;
	float a13;
	float a21;
	float a22;
	float a23;
	float a31;
	float a32;
	float a33;
};

/**
 * struct pe_coor_system - is a projected image of four-dim. axes in R**3
 * @axes[4]: coordinates of the x-, y-, z-, and w-axes in R**3
 * @origin : coordinates of the origin (0, 0, 0, 0) in R**3
 * @ave_dis: average distance between successive atoms of the protein backbone
 */
struct pe_coor_system {	
	struct pe_vector axes[4];		/* x, y, z, w */
	struct pe_vector origin;
	float ave_dis;
};

/**
 * struct pe_slt_tile - is a slant tile data
 * @cx        : x-coordinate of the center in R**3 (projected image)
 * @cy        : y-coordinate of the center in R**3 (projected image)
 * @cz        : z-coordinate of the center in R**3 (projected image)
 * @tx        : x-coordinate of the top vertex in R**3 (projected image)
 * @ty        : y-coordinate of the top vertex in R**3 (projected image)
 * @tz        : z-coordinate of the top vertex in R**3 (projected image)
 * @gx        : x-coordinate of the gradient in R**3 (projected image)
 * @gy        : y-coordinate of the gradient in R**3 (projected image)
 * @gz        : z-coordinate of the gradient in R**3 (projected image)
 * @dx        : x-coordinate of the direction to the target (projected image)
 * @dy        : y-coordinate of the direction to the target (projected image)
 * @dz        : z-coordinate of the direction to the target (projected image)
 * @error     : distance between the tile and the corresponding atom
 * @error_atom: the corresponding atom  (nearest atom)
 * @atom_flag : work flag
 * top[4]     : the coordinates of the top vertex in R**4
 * type       : type of the tile (XYZW, XZYW, ...)
 * code       : code of the tile (UP or DOWN)
 * ter        : flag for the end of a sequence
 */
struct pe_slt_tile {	
	float cx;		/* center of the tile */
	float cy;		/* center of the tile */
	float cz;		/* center of the tile */
	float tx;		/* top vertex of the tile */
	float ty;		/* top vertex of the tile */
	float tz;		/* top vertex of the tile */
	float bx;		/* bottom vertex of the tile */
	float by;		/* bottom vertex of the tile */
	float bz;		/* bottom vertex of the tile */
	float dx;		/* direction to the target */
	float dy;		/* direction to the target */
	float dz;		/* direction to the target */
	float error;		/* |tile - nearest_atom| */
	int error_atom;
	int atom_flag;
	int top[4];		/* top vertex in R**4 */
	int type;		/* XYZW, XZYW, ... */
	PE_CODE code;		/* C_UP or C_DOWN */
	int ter;		/* end of a sequence */
};

/**
 * struct pe_trn_table - is an entry of a lookup table of the encoding rule
 * @str[5]: type of the current tile ([xyz], [xzy], ...)
 * @cx    : x-coordinate of the center of the current tile
 * @cy    : y-coordinate of the center of the current tile
 * @cz    : z-coordinate of the center of the current tile
 * @cx    : x-coordinate of the direction of the current tile
 * @cy    : y-coordinate of the direction of the current tile
 * @cz    : z-coordinate of the direction of the current tile
 *
 * Note: origin = top vertex, i.e., top = ( 0, 0, 0 ) 
 */
struct pe_trn_table {
	char str[5];		/* [xyz], [xzy], ... */
	float cx;		/* center */
	float cy;		/* center */
	float cz;		/* center */
	float bx;                  /* bottom */
	float by;                  /* bottom */
	float bz;                  /* bottom */
	
	 /* downward */ 
	float downward_cx;	/* center */
	float downward_cy;
	float downward_cz;
	float downward_U_next_cx;	/* center of the next tile */
	float downward_U_next_cy;
	float downward_U_next_cz;
	float downward_U_tx;	/* top vertex */
	float downward_U_ty;
	float downward_U_tz;
	float downward_U_bx;	/* bottom vertex */
	float downward_U_by;
	float downward_U_bz;
	int downward_U_top[4];	/* top vertex in R**4 */
	int downward_U_type;	/* XYZW, XZYW, ...  */
	
	float downward_D_next_cx;	/* center of the next tile */
	float downward_D_next_cy;
	float downward_D_next_cz;
	float downward_D_tx;	/* top vertex */
	float downward_D_ty;
	float downward_D_tz;
	float downward_D_bx;	/* bottom vertex */
	float downward_D_by;
	float downward_D_bz;
	int downward_D_top[4];	/* top vertex in R**4 */
	int downward_D_type;	/* XYZW, XZYW, ...  */
	
	 /* upward */ 
	float upward_cx;	/* center */
	float upward_cy;
	float upward_cz;
	float upward_U_next_cx;	/* center of the next tile */
	float upward_U_next_cy;
	float upward_U_next_cz;
	float upward_U_tx;	/* top vertex */
	float upward_U_ty;
	float upward_U_tz;
	float upward_U_bx;	/* bottom vertex */
	float upward_U_by;
	float upward_U_bz;
	int upward_U_top[4];	/* top vertex in R**4 */
	int upward_U_type;	/* XYZW, XZYW, ...  */
	
	float upward_D_next_cx;	/* center of the next tile */
	float upward_D_next_cy;
	float upward_D_next_cz;
	float upward_D_tx;	/* top vertex */
	float upward_D_ty;
	float upward_D_tz;
	float upward_D_bx;	/* bottom vertex */
	float upward_D_by;
	float upward_D_bz;
	int upward_D_top[4];	/* top vertex in R**4 */
	int upward_D_type;	/* XYZW, XZYW, ...  */
};

/**
 * struct pe_error_table - is an entry of an approximation error table
 * @err: error of an approximation
 * @ind: the atom where the approximation started
 */
struct pe_error_table {
	float err;		/* error of the current approximation */
	int ind;		/* index of atoms */
};

/**
 * struct pe_path_of_length_n - is an entry of an approximation error table
 * @code: pointer to an UP/DOWN sequence of length n (including the current one) 
 * @c   : the position of the center of the n-th tile
 * @bb_c: pointer to the position of all tiles (from 0-th to (n-1)-th)
 * @type: the type of the (n-1)-th tile
 */
struct pe_path_of_length_n {
	PE_CODE * code;
	struct pe_vector c;
	struct pe_vector * bb_c;
	int type;
};

struct pe_init_tile {
	int ind;
	int t_ind;
	int t_type;
	float ave_error;
	float ave_dis;
};
