/*
 *  pe_initial_tile.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/10/06.
 *  Copyright 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern void print_trn_table(void);
extern void print_npath_table(void);
extern int seek_previous_target_atom(int ind);
extern int seek_next_target_atom(int ind);
extern void copy_slt_tile(struct pe_slt_tile * pt_dst_tile,
			  struct pe_slt_tile * pt_src_tile);
extern void move_tile_to_target(int t_ind, struct pe_slt_tile * slt_tiles,
				int atm_ct);
extern void rotate_tile_and_trn_table(int t_ind, struct pe_slt_tile * slt_tiles);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;
extern struct pe_trn_table my_trn_table[];
extern struct pe_coor_system my_coor;
extern int ston[];			/* pe_trn_table.c */
extern struct pe_qtos_table qtos[];	/* pe_trn_table.c */
extern struct pe_path_of_length_n ** npath_table;

void print_set_initial_tile(int ind,
			    float error_forward_up, float error_forward_down,
			    float error_backward_up, float error_backward_down,
			    int path_forward_up, int path_forward_down,
			    int path_backward_up, int path_backward_down,
			    int prv_ind, int nxt_ind)
{
	fprintf(STDOUT,
		"print_set_initial_tile() : %d-th atom --> %d-th atom --> %d-th atom\n",
		prv_ind, ind, nxt_ind);
	fprintf(STDOUT, "print_set_initial_tile() :       %d-th atm   = (%f, %f, %f)\n",
		prv_ind, my_protein.atom[prv_ind].x, my_protein.atom[prv_ind].y,
		my_protein.atom[prv_ind].z);
	fprintf(STDOUT, "print_set_initial_tile() :       %d-th atm   = (%f, %f, %f)\n",
		ind, my_protein.atom[ind].x, my_protein.atom[ind].y,
		my_protein.atom[ind].z);
	fprintf(STDOUT, "print_set_initial_tile() :       %d-th atm   = (%f, %f, %f)\n",
		nxt_ind, my_protein.atom[nxt_ind].x, my_protein.atom[nxt_ind].y,
		my_protein.atom[nxt_ind].z);
	if (error_backward_down > error_backward_up) {
		fprintf(STDOUT,
			"print_set_initial_tile() : DEBUG [ind = %d, downward] C_DOWN (error_backward_down = %f  >  error_backward_up = %f)\n",
			ind, sqrt(error_backward_down),sqrt(error_backward_up));
	} else {
		fprintf(STDOUT,
			"print_set_initial_tile() : DEBUG [ind = %d, downward] C_UP   (error_backward_down= %f < error_backward_up= %f)\n",
			ind, sqrt(error_backward_down),sqrt(error_backward_up));
	}
	fprintf(STDOUT,
		"print_set_initial_tile() :                                 (path_backward_down =   %d     path_backward_up = %d)\n",
		path_backward_down, path_backward_up);
	if (error_forward_down > error_forward_up) {
		fprintf(STDOUT,
			"print_set_initial_tile() : DEBUG [ind = %d, upward] C_UP   (error_forward_down= %f > error_forward_up= %f)\n",
			ind, sqrt(error_forward_down), sqrt(error_forward_up));
	} else {
		fprintf(STDOUT,
			"print_set_initial_tile() : DEBUG [ind = %d, upward] C_DOWN (error_forward_down = %f  <  error_forward_up = %f)\n",
			ind, sqrt(error_forward_down), sqrt(error_forward_up));
	}
	fprintf(STDOUT,
		"print_set_initial_tile() :                               (path_forward_down =   %d     path_forward_up = %d)\n",
		path_forward_down, path_forward_up);
	return;
}

float dist_from_next_of_downward_U_tile(int cur_ind,
					struct pe_slt_tile * wrk_tiles,
					int atm_ct)
{
	int cur_type;
	float cen_of_tile_x, cen_of_tile_y, cen_of_tile_z;
	float dist_x, dist_y, dist_z;
	float dist_squared;
	
	cur_type = wrk_tiles[cur_ind].type;
	cen_of_tile_x = my_trn_table[cur_type].downward_U_next_cx
		+ wrk_tiles[cur_ind].tx;
	cen_of_tile_y = my_trn_table[cur_type].downward_U_next_cy
		+ wrk_tiles[cur_ind].ty;
	cen_of_tile_z = my_trn_table[cur_type].downward_U_next_cz
		+ wrk_tiles[cur_ind].tz;
	dist_x = cen_of_tile_x -  my_protein.atom[atm_ct].x;
	dist_y = cen_of_tile_y -  my_protein.atom[atm_ct].y;
	dist_z = cen_of_tile_z -  my_protein.atom[atm_ct].z;
	dist_squared = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return dist_squared;
}

float dist_from_next_of_downward_D_tile(int cur_ind,
					struct pe_slt_tile * wrk_tiles,
					int atm_ct)
{
	int cur_type;
	float cen_of_tile_x, cen_of_tile_y, cen_of_tile_z;
	float dist_x, dist_y, dist_z;
	float dist_squared;
	
	cur_type = wrk_tiles[cur_ind].type;
	cen_of_tile_x = my_trn_table[cur_type].downward_D_next_cx
		+ wrk_tiles[cur_ind].tx;
	cen_of_tile_y = my_trn_table[cur_type].downward_D_next_cy
		+ wrk_tiles[cur_ind].ty;
	cen_of_tile_z = my_trn_table[cur_type].downward_D_next_cz
		+ wrk_tiles[cur_ind].tz;
	dist_x = cen_of_tile_x -  my_protein.atom[atm_ct].x;
	dist_y = cen_of_tile_y -  my_protein.atom[atm_ct].y;
	dist_z = cen_of_tile_z -  my_protein.atom[atm_ct].z;
	dist_squared = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return dist_squared;
}

float dist_from_next_of_upward_U_tile(int cur_ind,
				      struct pe_slt_tile * wrk_tiles,
				      int atm_ct)
{
	int cur_type;
	float cen_of_tile_x, cen_of_tile_y, cen_of_tile_z;
	float dist_x, dist_y, dist_z;
	float dist_squared;
	
	cur_type = wrk_tiles[cur_ind].type;
	cen_of_tile_x = my_trn_table[cur_type].upward_U_next_cx
		+ wrk_tiles[cur_ind].tx;
	cen_of_tile_y = my_trn_table[cur_type].upward_U_next_cy
		+ wrk_tiles[cur_ind].ty;
	cen_of_tile_z = my_trn_table[cur_type].upward_U_next_cz
		+ wrk_tiles[cur_ind].tz;
	dist_x = cen_of_tile_x -  my_protein.atom[atm_ct].x;
	dist_y = cen_of_tile_y -  my_protein.atom[atm_ct].y;
	dist_z = cen_of_tile_z -  my_protein.atom[atm_ct].z;
	dist_squared = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return dist_squared;
}

float dist_from_next_of_upward_D_tile(int cur_ind,
				      struct pe_slt_tile * wrk_tiles,
				      int atm_ct)
{
	int cur_type;
	float cen_of_tile_x, cen_of_tile_y, cen_of_tile_z;
	float dist_x, dist_y, dist_z;
	float dist_squared;
	
	cur_type = wrk_tiles[cur_ind].type;
	cen_of_tile_x = my_trn_table[cur_type].upward_D_next_cx
		+ wrk_tiles[cur_ind].tx;
	cen_of_tile_y = my_trn_table[cur_type].upward_D_next_cy
		+ wrk_tiles[cur_ind].ty;
	cen_of_tile_z = my_trn_table[cur_type].upward_D_next_cz
		+ wrk_tiles[cur_ind].tz;
	dist_x = cen_of_tile_x -  my_protein.atom[atm_ct].x;
	dist_y = cen_of_tile_y -  my_protein.atom[atm_ct].y;
	dist_z = cen_of_tile_z -  my_protein.atom[atm_ct].z;
	dist_squared = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return dist_squared;
}

/* Set the tile which corresponds to my_protein.atom[ind] to tile_type */
void set_tile_to(struct pe_slt_tile * slt_tiles, int ind, int tile_type)
{
	slt_tiles->type = tile_type;
	slt_tiles->ter = 0;
	slt_tiles->cx = my_protein.atom[ind].x;
	slt_tiles->cy = my_protein.atom[ind].y;
	slt_tiles->cz = my_protein.atom[ind].z;
	slt_tiles->tx = my_protein.atom[ind].x - my_trn_table[tile_type].cx;
	slt_tiles->ty = my_protein.atom[ind].y - my_trn_table[tile_type].cy;
	slt_tiles->tz = my_protein.atom[ind].z - my_trn_table[tile_type].cz;
	slt_tiles->bx = 0.0;
	slt_tiles->by = 0.0;
	slt_tiles->bz = 0.0;
	slt_tiles->error = (float) 0.0;
	slt_tiles->error_atom = ind;
	slt_tiles->atom_flag = 1;
	slt_tiles->top[0] = 0;
	slt_tiles->top[1] = 0;
	slt_tiles->top[2] = 0;
	slt_tiles->top[3] = 0;
	return;
}

float npath_inconsistency_check(struct pe_slt_tile * slt_tiles, int ind,
			       int path_forward_down, int path_forward_up,
			       float error_backward_down, 
			       float error_forward_down,
			       float error_backward_up,
			       float error_forward_up)
{
	float error_up, error_down;
	float bst_error;
	int bst_path;
	
	error_up = error_backward_up + error_forward_up;
	error_down = error_backward_down + error_forward_down;
	if (error_forward_down > error_forward_up) {
		bst_path = path_forward_up;
		bst_error = error_up;
	} else {
		bst_path = path_forward_down;
		bst_error = error_down;
	}
	if (slt_tiles->code == npath_table[XYZW][bst_path].code[0])
		return bst_error;
	
	fprintf(STDOUT,
		"npath_inconsistency_check() : Inconsistent initial code assgiment (ind= %d)\n",
		ind);
	if (error_up > error_down) {
		slt_tiles->code = C_DOWN;
		bst_error = error_down;
		if (0) fprintf(STDOUT,
			       "npath_inconsistency_check() : DEBUG set init. code to UP\n");
	} else {
		slt_tiles->code = C_UP;
		bst_error = error_up;
		if (0) fprintf(STDOUT,
			       "npath_inconsistency_check() : DEBUG set init. code to DOWN\n");
	}
	return bst_error;
}

float tile_inconsistency_check(struct pe_slt_tile * slt_tiles, int ind,
			      float downward_up, float upward_up,
			      float downward_down, float upward_down)
{
	float error_down, error_up;
	float bst_error;
	
	if (upward_up > upward_down) {
		error_down = downward_down * downward_down
				+ upward_down * upward_down;
		if (slt_tiles->code == C_DOWN)
			return error_down;
		fprintf(STDOUT,
			"tile_inconsistency_check(): Inconsistent initial code assgiment (ind= %d)\n",
			ind);
		error_up = downward_up * downward_up + upward_up * upward_up;
		if (error_down > error_up) {
			slt_tiles->code = C_UP;
			bst_error = error_up;
		} else {
			slt_tiles->code = C_DOWN;
			bst_error = error_down;
		}
	} else {
		error_up = downward_up * downward_up + upward_up * upward_up;
		if (slt_tiles->code == C_UP)
			return error_up;
		fprintf(STDOUT,
			"tile_inconsistency_check(): Inconsistent initial code assgiment (ind= %d)\n",
			ind);
		error_down = downward_down * downward_down
			+ upward_down * upward_down;
		if (error_down > error_up) {
			slt_tiles->code = C_UP;
			bst_error = error_up;
		} else {
			slt_tiles->code = C_DOWN;
			bst_error = error_down;
		}
	}
	if (0 && slt_tiles->code == C_UP)
		fprintf(STDOUT,
			"tile_inconsistency_check() : DEBUG set init. code to UP\n");
	else
		fprintf(STDOUT,
			"tile_inconsistency_check() : DEBUG set init. code to DOWN\n");
	return bst_error;
}

int compute_cur_tile_type(int t_ind, struct pe_slt_tile * slt_tiles,
			  int cur_code, PE_DIRECTION direction)
{
	int cur_type;
	int nxt_ind, nxt_type;
	int prv_ind, prv_type;
	PE_CODE nxt_code, prv_code;
	
	if (slt_tiles[t_ind].type < 0) {	/* if not assigned */
		if (direction == DIR_FORWARD) {
			prv_ind = t_ind - 1;
			prv_type = slt_tiles[prv_ind].type;
			prv_code = slt_tiles[prv_ind].code;
			if (prv_code == C_UP && cur_code == C_DOWN)
				cur_type = my_trn_table[prv_type].upward_D_type;
			else if (prv_code == C_DOWN && cur_code == C_DOWN)
				cur_type = 
					my_trn_table[prv_type].downward_D_type;
			else if (prv_code == C_UP && cur_code == C_UP)
				cur_type = my_trn_table[prv_type].upward_U_type;
			else
				cur_type = 
					my_trn_table[prv_type].downward_U_type;
		} else {
			nxt_ind = t_ind + 1;
			nxt_type = slt_tiles[nxt_ind].type;
			nxt_code = slt_tiles[nxt_ind].code;
			if (cur_code == C_UP && nxt_code == C_DOWN)
				cur_type = my_trn_table[nxt_type].upward_D_type;
			else if (cur_code == C_UP && nxt_code == C_UP)
				cur_type =
					my_trn_table[nxt_type].downward_D_type;
			else if (cur_code == C_DOWN && nxt_code == C_DOWN)
				cur_type = my_trn_table[nxt_type].upward_U_type;
			else
				cur_type =
					my_trn_table[nxt_type].downward_U_type;
		}
	} else {
		cur_type = slt_tiles[t_ind].type;
	}
	return cur_type;
}

float dist_from_npath(int t_ind, struct pe_slt_tile * slt_tiles,
		      int cur_type, int path_ct, int tgt_atm)
{
	float npath_offsetx, npath_offsety, npath_offsetz;
	float dist_x, dist_y, dist_z;
	float squared_length;
	
	npath_offsetx = slt_tiles[t_ind].cx - my_trn_table[cur_type].cx;
	npath_offsety = slt_tiles[t_ind].cy - my_trn_table[cur_type].cy;
	npath_offsetz = slt_tiles[t_ind].cz - my_trn_table[cur_type].cz;
	dist_x = (npath_table[cur_type][path_ct].c.x + npath_offsetx)
		- my_protein.atom[tgt_atm].x;
	dist_y = (npath_table[cur_type][path_ct].c.y + npath_offsety)
		- my_protein.atom[tgt_atm].y;
	dist_z = (npath_table[cur_type][path_ct].c.z + npath_offsetz)
		- my_protein.atom[tgt_atm].z;
	squared_length = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return squared_length;
}

float dist_bb_from_npath(int t_ind, struct pe_slt_tile * slt_tiles, 
			 int cur_type, int path_ct, int tile_ct,
			 int tgt_atm)
{
	float npath_offsetx, npath_offsety, npath_offsetz;
	float dist_x, dist_y, dist_z;
	float squared_length;
	
	npath_offsetx = slt_tiles[t_ind].cx - my_trn_table[cur_type].cx;
	npath_offsety = slt_tiles[t_ind].cy - my_trn_table[cur_type].cy;
	npath_offsetz = slt_tiles[t_ind].cz - my_trn_table[cur_type].cz;
	dist_x = (npath_table[cur_type][path_ct].bb_c[tile_ct].x
		  + npath_offsetx) - my_protein.atom[tgt_atm].x;
	dist_y = (npath_table[cur_type][path_ct].bb_c[tile_ct].y
		  + npath_offsety) - my_protein.atom[tgt_atm].y;
	dist_z = (npath_table[cur_type][path_ct].bb_c[tile_ct].z
		  + npath_offsetz) - my_protein.atom[tgt_atm].z;
	squared_length = dist_x * dist_x + dist_y * dist_y + dist_z * dist_z;
	return squared_length;
}

/* Set center of the next ``flat'' tile [backward] */
void set_cen_of_new_tile_back(int new_ind, struct pe_slt_tile * wrk_tiles,
			      int cur_ind)
{
	int cur_type;
	
	cur_type = wrk_tiles[cur_ind].type;
	if (wrk_tiles[cur_ind].code == C_UP) {	/*  <- U */
		wrk_tiles[new_ind].cx = my_trn_table[cur_type].downward_cx
			+ wrk_tiles[cur_ind].tx;
		wrk_tiles[new_ind].cy = my_trn_table[cur_type].downward_cy
			+ wrk_tiles[cur_ind].ty;
		wrk_tiles[new_ind].cz = my_trn_table[cur_type].downward_cz
			+ wrk_tiles[cur_ind].tz;
	} else if (wrk_tiles[cur_ind].code == C_DOWN) {	/*  <- D */
		wrk_tiles[new_ind].cx = my_trn_table[cur_type].upward_cx
			+ wrk_tiles[cur_ind].tx;
		wrk_tiles[new_ind].cy = my_trn_table[cur_type].upward_cy
			+ wrk_tiles[cur_ind].ty;
		wrk_tiles[new_ind].cz = my_trn_table[cur_type].upward_cz
			+ wrk_tiles[cur_ind].tz;
	} else {
		fprintf(STDERR,
			"set_cen_of_new_tile_back() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}
	if (DEBUG) {
		float wrk_x, wrk_y, wrk_z;
		float wrk_length;
	
		wrk_x = (wrk_tiles[cur_ind].cx - wrk_tiles[cur_ind].tx) - my_trn_table[cur_type].cx;
		wrk_y = (wrk_tiles[cur_ind].cy - wrk_tiles[cur_ind].ty) - my_trn_table[cur_type].cy;
		wrk_z = (wrk_tiles[cur_ind].cz - wrk_tiles[cur_ind].tz) - my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"set_cen_of_new_tile_back() : ERROR |wrk_tiles[%d].(c* - t*) - my_trn_table[%d].c*| = %8.3f\n",
				cur_ind, cur_type, wrk_length);
			print_tile(cur_ind, wrk_tiles);
			fprintf(STDERR,
				"set_cen_of_new_tile_back() : ERROR my_trn_table[%d].c*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].cx,
				my_trn_table[cur_type].cy, my_trn_table[cur_type].cz);
			if (0) print_trn_table();
			if (0) print_my_slt_tiles(wrk_tiles);
			exit(1);
		}
		if (my_settings.ca_only_flag == FALSE) {
			wrk_x = (wrk_tiles[cur_ind].bx - wrk_tiles[cur_ind].tx) - my_trn_table[cur_type].bx;
			wrk_y = (wrk_tiles[cur_ind].by - wrk_tiles[cur_ind].ty) - my_trn_table[cur_type].by;
			wrk_z = (wrk_tiles[cur_ind].bz - wrk_tiles[cur_ind].tz) - my_trn_table[cur_type].bz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_cen_of_new_tile_back() : ERROR |wrk_tiles[%d].(b* - t*) - my_trn_table[%d].b*| = %8.3f\n",
					cur_ind, cur_type, wrk_length);
				print_tile(cur_ind, wrk_tiles);
				fprintf(STDERR,
					"set_cen_of_new_tile_back() : ERROR my_trn_table[%d].b*= (%8.3f, %8.3f, %8.3f)\n",
					cur_type, my_trn_table[cur_type].bx,
					my_trn_table[cur_type].by, my_trn_table[cur_type].bz);
				if (0) print_trn_table();
				exit(1);
			}
		}
	}
	if (DEBUG) {
		float wrk_x, wrk_y, wrk_z;
		float wrk_length, wrk2_length;
	
		wrk_x = wrk_tiles[new_ind].cx - wrk_tiles[cur_ind].cx;
		wrk_y = wrk_tiles[new_ind].cy - wrk_tiles[cur_ind].cy;
		wrk_z = wrk_tiles[new_ind].cz - wrk_tiles[cur_ind].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (wrk_tiles[cur_ind].code == C_UP) {
			wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].cx;
			wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].cy;
			wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].cz;
		} else {
			wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].cx;
			wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].cy;
			wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].cz;
		}
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"set_cen_of_new_tile_back() : wrk_length= %f <=> wrk2_length= %f from %d-th to %d-th tile\n",
				wrk_length, wrk2_length, cur_ind, new_ind);
			exit(1);
		}
	}
	return;
}

/* Set center of the next ``flat'' tile [backward] */
void set_cen_of_new_tile_for(int new_ind, struct pe_slt_tile * wrk_tiles, 
			     int cur_ind)
{
	int cur_type;
	
	cur_type = wrk_tiles[cur_ind].type;
	if (wrk_tiles[cur_ind].code == C_DOWN) {	/* D ->  */
		wrk_tiles[new_ind].cx = my_trn_table[cur_type].downward_cx
			+ wrk_tiles[cur_ind].tx;
		wrk_tiles[new_ind].cy = my_trn_table[cur_type].downward_cy
			+ wrk_tiles[cur_ind].ty;
		wrk_tiles[new_ind].cz = my_trn_table[cur_type].downward_cz
			+ wrk_tiles[cur_ind].tz;
	} else if (wrk_tiles[cur_ind].code == C_UP) {	/* U ->  */
		wrk_tiles[new_ind].cx = my_trn_table[cur_type].upward_cx
		+ wrk_tiles[cur_ind].tx;
		wrk_tiles[new_ind].cy = my_trn_table[cur_type].upward_cy
			+ wrk_tiles[cur_ind].ty;
		wrk_tiles[new_ind].cz = my_trn_table[cur_type].upward_cz
			+ wrk_tiles[cur_ind].tz;
	} else {
		fprintf(STDERR,
			"set_cen_of_new_tile_for() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}

	if (DEBUG) {
		float wrk_x, wrk_y, wrk_z;
		float wrk_length;
	
		wrk_x = (wrk_tiles[cur_ind].cx - wrk_tiles[cur_ind].tx)
			- my_trn_table[cur_type].cx;
		wrk_y = (wrk_tiles[cur_ind].cy - wrk_tiles[cur_ind].ty)
			- my_trn_table[cur_type].cy;
		wrk_z = (wrk_tiles[cur_ind].cz - wrk_tiles[cur_ind].tz)
			- my_trn_table[cur_type].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length, 1.0)) {
			fprintf(STDERR,
				"set_cen_of_new_tile_for() : ERROR |wrk_tiles[%d].(c* - t*) - my_trn_table[%d].c*| = %8.3f\n",
				cur_ind, cur_type, wrk_length);
			print_tile(cur_ind, wrk_tiles);
			fprintf(STDERR,
				"set_cen_of_new_tile_for() : ERROR my_trn_table[%d].c*= (%8.3f, %8.3f, %8.3f)\n",
				cur_type, my_trn_table[cur_type].cx,
				my_trn_table[cur_type].cy, my_trn_table[cur_type].cz);
			if (0) print_trn_table();
			if (0) print_my_slt_tiles(wrk_tiles);
			exit(1);
		}
		if (my_settings.ca_only_flag == FALSE) {
			wrk_x = (wrk_tiles[cur_ind].bx - wrk_tiles[cur_ind].tx)
			- my_trn_table[cur_type].bx;
			wrk_y = (wrk_tiles[cur_ind].by - wrk_tiles[cur_ind].ty)
				- my_trn_table[cur_type].by;
			wrk_z = (wrk_tiles[cur_ind].bz - wrk_tiles[cur_ind].tz)
				- my_trn_table[cur_type].bz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_cen_of_new_tile_for() : ERROR |wrk_tiles[%d].(b* - t*) - my_trn_table[%d].b*| = %8.3f\n",
					cur_ind, cur_type, wrk_length);
				print_tile(cur_ind, wrk_tiles);
				fprintf(STDERR,
					"set_cen_of_new_tile_for() : ERROR my_trn_table[%d].b*= (%8.3f, %8.3f, %8.3f)\n",
					cur_type, my_trn_table[cur_type].bx,
					my_trn_table[cur_type].by, my_trn_table[cur_type].bz);
				if (0) print_trn_table();
				exit(1);
			}
		}
	}

	if (DEBUG) {
		float wrk_x, wrk_y, wrk_z;
		float wrk_length, wrk2_length;
	
		wrk_x = wrk_tiles[new_ind].cx - wrk_tiles[cur_ind].cx;
		wrk_y = wrk_tiles[new_ind].cy - wrk_tiles[cur_ind].cy;
		wrk_z = wrk_tiles[new_ind].cz - wrk_tiles[cur_ind].cz;
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (wrk_tiles[cur_ind].code == C_DOWN) {
			wrk_x = my_trn_table[cur_type].downward_cx - my_trn_table[cur_type].cx;
			wrk_y = my_trn_table[cur_type].downward_cy - my_trn_table[cur_type].cy;
			wrk_z = my_trn_table[cur_type].downward_cz - my_trn_table[cur_type].cz;
		} else {
			wrk_x = my_trn_table[cur_type].upward_cx - my_trn_table[cur_type].cx;
			wrk_y = my_trn_table[cur_type].upward_cy - my_trn_table[cur_type].cy;
			wrk_z = my_trn_table[cur_type].upward_cz - my_trn_table[cur_type].cz;
		}
		wrk2_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk2_length, 1.0)) {
			fprintf(STDERR,
				"set_cen_of_new_tile_for() : wrk_length= %f <=> wrk2_length= %f from %d-th to %d-th tile\n",
				wrk_length, wrk2_length, cur_ind, new_ind);
			exit(1);
		}
	}
	return;
}

/* backward */
void set_top_of_new_tile_back(int new_ind, struct pe_slt_tile * wrk_tiles,
			      PE_CODE new_code, int cur_ind)
{
	int cur_type;
	int i;
	
	cur_type = wrk_tiles[cur_ind].type;
	if (wrk_tiles[cur_ind].code == C_UP) {
		if (new_code == C_UP) {	/* U <- U */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].downward_D_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].downward_D_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].downward_D_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_UP;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].downward_D_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].downward_D_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].downward_D_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].downward_D_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].downward_D_type;
		} else {	/* D <- U */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].downward_U_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].downward_U_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].downward_U_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_DOWN;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].downward_U_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].downward_U_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].downward_U_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].downward_U_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].downward_U_type;
		}
	} else if (wrk_tiles[cur_ind].code == C_DOWN) {
		if (new_code == C_UP) {	/* U <- D */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].upward_D_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].upward_D_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].upward_D_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_UP;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].upward_D_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].upward_D_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].upward_D_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].upward_D_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].upward_D_type;
		} else {	/* D <- D */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].upward_U_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].upward_U_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].upward_U_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_DOWN;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].upward_U_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].upward_U_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].upward_U_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].upward_U_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].upward_U_type;
		}
	} else {
		fprintf(STDERR,
			"set_top_of_new_tile_back() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}

	if (DEBUG) {
		float wrk_x, wrk_y, wrk_z;
		float wrk_length, wrk_length2;
	
		wrk_x = (wrk_tiles[new_ind].cx - wrk_tiles[new_ind].tx);
		wrk_y = (wrk_tiles[new_ind].cy - wrk_tiles[new_ind].ty);
		wrk_z = (wrk_tiles[new_ind].cz - wrk_tiles[new_ind].tz);
		wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		wrk_x = (wrk_tiles[new_ind].cx - wrk_tiles[new_ind].bx);
		wrk_y = (wrk_tiles[new_ind].cy - wrk_tiles[new_ind].by);
		wrk_z = (wrk_tiles[new_ind].cz - wrk_tiles[new_ind].bz);
		wrk_length2 = sqrt(wrk_x * wrk_x + wrk_y * wrk_y
				  + wrk_z * wrk_z);
		if (FLOAT_NEQ(wrk_length - wrk_length2, 1.0)) {
			fprintf(STDERR,
				"set_top_of_new_tile_back() : DEBUG wrk_tiles[%d] |cen - top|= %f\n",
				new_ind, wrk_length);
			fprintf(STDERR,
				"set_top_of_new_tile_back() : DEBUG wrk_tiles[%d] |cen - bot|= %f\n",
				new_ind, wrk_length2);
			exit(1);
		}
	}
	return;
}

/* Forward */
void set_top_of_new_tile_for(int new_ind, struct pe_slt_tile * wrk_tiles,
			     PE_CODE new_code, int cur_ind)
{
	int cur_type;
	int i;
	
	cur_type = wrk_tiles[cur_ind].type;
	if (wrk_tiles[cur_ind].code == C_DOWN) {
		if (new_code == C_DOWN) {	/* D -> D */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].downward_D_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].downward_D_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].downward_D_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_DOWN;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].downward_D_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].downward_D_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].downward_D_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].downward_D_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].downward_D_type;
		} else {	/* D -> U */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].downward_U_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].downward_U_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].downward_U_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_UP;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].downward_U_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].downward_U_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].downward_U_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].downward_U_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].downward_U_type;
		}
	} else if (wrk_tiles[cur_ind].code == C_UP) {
		if (new_code == C_DOWN) {	/* U -> D */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].upward_D_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].upward_D_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].upward_D_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_DOWN;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].upward_D_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].upward_D_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].upward_D_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].upward_D_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].upward_D_type;
		} else {	/* U -> U */
			wrk_tiles[new_ind].tx
				= my_trn_table[cur_type].upward_U_tx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].ty
				= my_trn_table[cur_type].upward_U_ty
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].tz
				= my_trn_table[cur_type].upward_U_tz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].code = C_UP;
			for (i = 0; i < 4; i++)
				wrk_tiles[new_ind].top[i]
				    = my_trn_table[cur_type].upward_U_top[i]
				    + wrk_tiles[cur_ind].top[i];
			wrk_tiles[new_ind].bx
				= my_trn_table[cur_type].upward_U_bx
				+ wrk_tiles[cur_ind].tx;
			wrk_tiles[new_ind].by
				= my_trn_table[cur_type].upward_U_by
				+ wrk_tiles[cur_ind].ty;
			wrk_tiles[new_ind].bz
				= my_trn_table[cur_type].upward_U_bz
				+ wrk_tiles[cur_ind].tz;
			wrk_tiles[new_ind].type
				= my_trn_table[cur_type].upward_U_type;
		}
	} else {
		fprintf(STDERR,
			"set_top_of_new_tile_for() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}
	return;
}

/* distance between the current tile and the target atom */
float dist_from_cur_tile(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct)
{
	float tgt_atm_x, tgt_atm_y, tgt_atm_z;
	float tile_x, tile_y, tile_z;
	float dist_squared;
	
	tgt_atm_x = my_protein.atom[atm_ct].x;
	tgt_atm_y = my_protein.atom[atm_ct].y;
	tgt_atm_z = my_protein.atom[atm_ct].z;
	tile_x = my_trn_table[slt_tiles[t_ind].type].cx + slt_tiles[t_ind].tx;
	tile_y = my_trn_table[slt_tiles[t_ind].type].cy + slt_tiles[t_ind].ty;
	tile_z = my_trn_table[slt_tiles[t_ind].type].cz + slt_tiles[t_ind].tz;
	dist_squared = (tile_x - tgt_atm_x) * (tile_x - tgt_atm_x)
		+ (tile_y - tgt_atm_y) * (tile_y - tgt_atm_y)
		+ (tile_z - tgt_atm_z) * (tile_z - tgt_atm_z);
	return dist_squared;
}

void setup_loop6_table(int t_ind, struct pe_slt_tile * slt_tiles,
		       struct pe_slt_tile * loop6_tiles, float * loop6_error,
		       float cur_error)
{
	int i;
	
	copy_slt_tile(loop6_tiles, slt_tiles + t_ind);
	loop6_error[0] = cur_error;
	for (i = 0; i < 6; i++) {
		loop6_tiles[i].ter = 0;
		loop6_error[i] = 0.0;
	}
	return;
}

int compute_new_toggle6_index(int old_ind, PE_DIRECTION direction)
{
	int new_ind;
	if (direction == DIR_BACKWARD) {
		if (old_ind <= 0)
			new_ind = 5;
		else
			new_ind = old_ind - 1;
	} else {
		if (old_ind >= 5)
			new_ind = 0;
		else
			new_ind = old_ind + 1;
	}
	return new_ind;
}

/* Seek minimum in loop6_error[] */
int choose_best_of_loop6_tiles(int t_ind, int last6_ind, int atm_ct,
			       float * loop6_error, float next_error)
{
	int i;
	int min_ind;
	float min_error;
	
	fprintf(STDERR,
		"choose_best_of_loop6_tiles() : CAUTION 6-tile loop check (t_ind=%d, atm_ct=%d)\n",
		t_ind, atm_ct);
	fprintf(STDERR,
		"choose_best_of_loop6_tiles() :         (loop6_error[%d]=%f) <==> (next_error=%f)\n",
		last6_ind, loop6_error[last6_ind], next_error);
	for (i = 0; i < 6; i++) {
		fprintf(STDERR,
			"choose_best_of_loop6_tiles() :                  loop6_error[%d]= %8.3f\n",
			i, loop6_error[i]);
	}
	
	min_ind = 0;
	min_error = loop6_error[0];
	for (i = 1; i < 6; i++) {
		if (min_error > loop6_error[i]) {
			min_ind = i;
			min_error = loop6_error[i];
		}
	}
	return min_ind;
}

int seek_init_tile_overflow_check(int while_ct, int last6_ind,
				   float * loop6_error, float cur_error,
				   float next_error)
{
	if (while_ct > my_protein.tile_max_num - 20) {
		fprintf(STDERR,
			"seek_init_tile_overflow_check() : cur_error= %f, next_error= %f <==> loop6_error[%d]=%f  (while_ct= %d)\n",
			cur_error, next_error, last6_ind,
			loop6_error[last6_ind], while_ct);
		if (!FLOAT_NEQ(next_error - loop6_error[last6_ind],
			       my_coor.ave_dis)) {
			exit(1);
		}
		if (while_ct > my_protein.tile_max_num) {
			fprintf(STDERR,
				"seek_init_tile_overflow_check() : ERROR  overflow (while_ct=%d) > (my_protein.tile_max_num=%d)\n",
				while_ct, my_protein.tile_max_num);
			fprintf(STDERR,
				"seek_init_tile_overflow_check() :         cur_error= %8.3f, next_error= %8.3f\n",
				cur_error, next_error);
			if (cur_error < next_error) {
			return FAIL;
			}
		}
	}
	return SUCCESS;
}

/* Set the initial tile */
int copy_init_tile(int t_ind, struct pe_slt_tile * slt_tiles,
		   int toggle6_ind, struct pe_slt_tile * loop6_tiles,
		   PE_DIRECTION direction)
{
	int new_ind;
	
	if (direction == DIR_BACKWARD)
		new_ind = t_ind - 1;
	else
		new_ind = t_ind + 1;
	
	copy_slt_tile(slt_tiles + new_ind, loop6_tiles + toggle6_ind);
	my_protein.tile_num++;
	return new_ind;
}

/* Return "distance squared" to save the computation of sqrt() */
/* Backward: CA(nxt_ind) <-C(nxt_ind+1)<-N(nxt_ind+2)<-  CA(cur_atm) */
/* Note that the C and N atoms may not exist in the file. */
float backward_backbone_dist_check(int t_ind, struct pe_slt_tile * slt_tiles,
				   int cur_type, int path_ct, int nxt_ind)
{
	int tgt_ind;
	int tile_ct;
	float wrk_length, bst_length;
	float sum_squared_length;
	
	tgt_ind = nxt_ind + 1;
	if (my_protein.atom[tgt_ind].atom_name == PE_CC) {
		tile_ct = my_settings.path_length - 1;
		bst_length = dist_bb_from_npath(t_ind, slt_tiles, cur_type,
						path_ct, tile_ct, tgt_ind);
		wrk_length = bst_length;
		tile_ct--;
		while (tile_ct > 0) {
			wrk_length = dist_bb_from_npath(t_ind, slt_tiles,
							cur_type, path_ct,
							tile_ct,tgt_ind);
			if (bst_length < wrk_length)
				break;
			bst_length = wrk_length;
			tile_ct--;
		}
		tgt_ind++;
	} else {
		bst_length = 0.0;
	}
	sum_squared_length = bst_length;	/* C */
	
	if (my_protein.atom[tgt_ind].atom_name == PE_NN) {
		tile_ct = 1;
		bst_length = dist_bb_from_npath(t_ind, slt_tiles, cur_type,
						path_ct, tile_ct, tgt_ind);
		tile_ct++;
		while (tile_ct < my_settings.path_length) {
			wrk_length = dist_bb_from_npath(t_ind, slt_tiles,
							cur_type, path_ct,
							tile_ct, tgt_ind);
			if (bst_length < wrk_length)
				break;
			bst_length = wrk_length;
			tile_ct++;
		}
	} else {
		bst_length = 0.0;
	}
	sum_squared_length = sum_squared_length + bst_length;	/* C + N */
	return sum_squared_length;
}

/* Return "distance squared" to save the computation of sqrt() */
/* Forward: CA(cur_ind) ->C(nxt_ind-2)-> N(nxt_ind-1)-> CA(nxt_ind) */
/* Note that the C and N atoms may not exist in the file. */
float forward_backbone_dist_check(int t_ind, struct pe_slt_tile * slt_tiles,
				  int cur_type, int path_ct, int nxt_ind)
{
	int tgt_ind;
	int tile_ct;
	float wrk_length, bst_length;
	float sum_squared_length;
	
	tgt_ind = nxt_ind - 1;
	if (my_protein.atom[tgt_ind].atom_name == PE_NN) {
		tile_ct = my_settings.path_length - 1;
		bst_length = dist_bb_from_npath(t_ind, slt_tiles, cur_type,
						path_ct, tile_ct, tgt_ind);
		wrk_length = bst_length;
		tile_ct--;
		while (tile_ct > 0) {
			wrk_length = dist_bb_from_npath(t_ind, slt_tiles,
							cur_type, path_ct,
							tile_ct,tgt_ind);
			if (bst_length < wrk_length)
				break;
			bst_length = wrk_length;
			tile_ct--;
		}
		tgt_ind--;
	} else {
		bst_length = 0.0;
	}
	sum_squared_length = bst_length;	/* N */
	
	if (my_protein.atom[tgt_ind].atom_name == PE_CC) {
		tile_ct = 1;
		bst_length = dist_bb_from_npath(t_ind, slt_tiles, cur_type,
						path_ct, tile_ct, tgt_ind);
		tile_ct++;
		while (tile_ct < my_settings.path_length) {
			wrk_length = dist_bb_from_npath(t_ind, slt_tiles,
							cur_type, path_ct,
							tile_ct, tgt_ind);
			if (bst_length < wrk_length)
				break;
			bst_length = wrk_length;
			tile_ct++;
		}
	} else {
		bst_length = 0.0;
	}
	sum_squared_length = sum_squared_length + bst_length;	/* C + N */
	return sum_squared_length;
}

/* bst_length1 = better_dist_from_npath(t_ind, slt_tiles,
C_DOWN, atm_ct, direction, &bst_path1, &bst_type1);
*/
float better_dist_from_npath(int t_ind, struct pe_slt_tile * slt_tiles,
			     PE_CODE cur_code, int atm_ct,
			     PE_DIRECTION direction, int * pbst_path,
			     int * pbst_type)
{
	int max_path;
	int cur_type;
	int p_ct, min_p_ct, max_p_ct;
	float bst_length, wrk_length, bb_bst_length;
	int bst_path;
	
	max_path = my_settings.max_path;
	cur_type = compute_cur_tile_type(t_ind, slt_tiles, cur_code, direction);
	bst_length = -1.0;
	bst_path = -1;
	if ( (cur_code == C_DOWN && direction == DIR_FORWARD)
	     || (cur_code == C_UP && direction == DIR_BACKWARD)) {
		min_p_ct = 0;
		max_p_ct =  max_path / 2;
	} else {
		min_p_ct = max_path / 2;
		max_p_ct =  max_path;
	}
	for (p_ct = min_p_ct; p_ct < max_p_ct; p_ct++) {
		wrk_length = dist_from_npath(t_ind, slt_tiles, cur_type, p_ct,
					     atm_ct);
		if (my_settings.along_backbone == TRUE) {
			if (direction == DIR_FORWARD) {
				bb_bst_length
				= forward_backbone_dist_check(t_ind, slt_tiles,
							      cur_type, p_ct,
							      atm_ct);
			} else {
				bb_bst_length
				= backward_backbone_dist_check(t_ind, slt_tiles,
							       cur_type, p_ct,
							       atm_ct);
			}
			wrk_length = wrk_length + bb_bst_length;
		}
		if (bst_length > wrk_length || bst_length < 0.0) {
			bst_length = wrk_length;
			bst_path = p_ct;
		}
	}
	*pbst_path = bst_path;
	*pbst_type = cur_type;
	return bst_length;
}

float dist_from_new_tile(int t_ind, struct pe_slt_tile * slt_tiles,
			 PE_CODE cur_code, int new_ind, PE_DIRECTION direction)
{
	int cur_type;
	float dist_squared;
	float centre_of_tile_x, centre_of_tile_y, centre_of_tile_z;
	float dist_x, dist_y, dist_z;
	
	cur_type = compute_cur_tile_type(t_ind, slt_tiles, cur_code, direction);
	if ((cur_code == C_DOWN && direction == DIR_BACKWARD)
	    || (cur_code == C_UP && direction == DIR_FORWARD)) {
		centre_of_tile_x = my_trn_table[cur_type].upward_cx
		+ slt_tiles[t_ind].tx;
		centre_of_tile_y = my_trn_table[cur_type].upward_cy
			+ slt_tiles[t_ind].ty;
		centre_of_tile_z = my_trn_table[cur_type].upward_cz
			+ slt_tiles[t_ind].tz;
	} else {
		centre_of_tile_x = my_trn_table[cur_type].downward_cx
		+ slt_tiles[t_ind].tx;
		centre_of_tile_y = my_trn_table[cur_type].downward_cy
			+ slt_tiles[t_ind].ty;
		centre_of_tile_z = my_trn_table[cur_type].downward_cz
			+ slt_tiles[t_ind].tz;
	}
	dist_x = centre_of_tile_x - my_protein.atom[new_ind].x;
	dist_y = centre_of_tile_y - my_protein.atom[new_ind].y;
	dist_z = centre_of_tile_z - my_protein.atom[new_ind].z;
	dist_squared = dist_x*dist_x + dist_y*dist_y + dist_z*dist_z;
	return dist_squared;
}

void dist_from_next_tile_back(int cur_ind, struct pe_slt_tile * wrk_tiles,
			      int atm_ct, float * pnxt_U_error,
			      float * pnxt_D_error)
{
	
	if (wrk_tiles[cur_ind].code == C_UP) {
		*pnxt_U_error = dist_from_next_of_downward_U_tile(cur_ind,
							wrk_tiles, atm_ct);
		*pnxt_D_error = dist_from_next_of_downward_D_tile(cur_ind,
							wrk_tiles, atm_ct);
	} else if (wrk_tiles[cur_ind].code == C_DOWN) {
		*pnxt_U_error = dist_from_next_of_upward_U_tile(cur_ind,
							wrk_tiles, atm_ct);
		*pnxt_D_error = dist_from_next_of_upward_D_tile(cur_ind,
							wrk_tiles, atm_ct);
	} else {
		fprintf(STDERR,
			"dist_from_next_tile_back() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}
	return;
}

void dist_from_next_tile_for(int cur_ind, struct pe_slt_tile * wrk_tiles,
			     int atm_ct, float * pnxt_U_error,
			     float * pnxt_D_error)
{	
	if (wrk_tiles[cur_ind].code == C_DOWN) {
		*pnxt_U_error = dist_from_next_of_downward_U_tile(cur_ind,
							wrk_tiles, atm_ct);
		*pnxt_D_error = dist_from_next_of_downward_D_tile(cur_ind,
							wrk_tiles, atm_ct);
	} else if (wrk_tiles[cur_ind].code == C_UP) {
		*pnxt_U_error = dist_from_next_of_upward_U_tile(cur_ind,
							wrk_tiles, atm_ct);
		*pnxt_D_error = dist_from_next_of_upward_D_tile(cur_ind,
							wrk_tiles, atm_ct);
	} else {
		fprintf(STDERR,
			"dist_from_next_tile_for() : ERROR wrk_tiles[%d].code == UNKOWN\n",
			cur_ind);
		exit(1);
	}
	return;
}

/* There are 4 types of gradient: [pqr], [tpq], [stp], qnd [rst]. */
int best_init_tile_and_code(int nxt_t_ind, struct pe_slt_tile * slt_tiles,
			    int atm_ct, PE_DIRECTION direction,
			    PE_CODE * pbst_code, int * pbst_type)
{
	float bst_error, wrk_error;
	int bst_path, bst_type;
	int wrk_path, wrk_type;
	int bst_grad_type, grad_type;
	PE_CODE bst_code;
	int max_path;
	int new_atm;
	int cur_type, cur_type_saved;
	
	cur_type_saved = slt_tiles[nxt_t_ind].type;
	bst_path = -1;
	bst_type = -1;
	bst_grad_type = -1;
	bst_code = -1;
	bst_error = -1.0;
	if (direction == DIR_BACKWARD) {
		new_atm = seek_previous_target_atom(atm_ct);
		if (new_atm <= 0) {
			fprintf(STDERR,
				"best_init_tile_and_code() : There exists no CA atom [backward] (FROM %d TO %d).\n",
				new_atm + 1, atm_ct - 1);
			new_atm = atm_ct - 1;
		}
	} else {
		new_atm = seek_next_target_atom(atm_ct);
		if (new_atm > my_protein.atm_num) {
			fprintf(STDERR,
				"best_init_tile_and_code() : There exists no CA atom [forward] (FROM %d TO %d).\n",
				atm_ct + 1, new_atm - 1);
			new_atm = atm_ct + 1;
		}
	}
	
	cur_type = slt_tiles[nxt_t_ind].type;
	for (grad_type = 0; grad_type < 4; grad_type++) {
		slt_tiles[nxt_t_ind].type = cur_type;
		if (my_settings.ca_only_flag == TRUE) {
			wrk_error = better_dist_from_npath(nxt_t_ind, slt_tiles,
						C_DOWN, new_atm, direction,
						&wrk_path, &wrk_type);
		} else {
			slt_tiles[nxt_t_ind].tx = slt_tiles[nxt_t_ind].cx
			- my_trn_table[cur_type].cx;
			slt_tiles[nxt_t_ind].ty = slt_tiles[nxt_t_ind].cy
				- my_trn_table[cur_type].cy;
			slt_tiles[nxt_t_ind].tz = slt_tiles[nxt_t_ind].cz
				- my_trn_table[cur_type].cz;
			wrk_error = dist_from_new_tile(nxt_t_ind, slt_tiles,
						C_DOWN, new_atm, direction);
		}
		if ((bst_error > wrk_error) || (bst_error < 0.0)) {
			bst_error = wrk_error;
			bst_grad_type = grad_type;
			if (my_settings.ca_only_flag == TRUE) {
				bst_path = wrk_path;
				bst_type = wrk_type;
			} else {
				bst_code = C_DOWN;
				bst_type = cur_type;
			}
		}
		
		if (my_settings.ca_only_flag == TRUE) {
			wrk_error = better_dist_from_npath(nxt_t_ind, slt_tiles,
						C_UP, new_atm, direction,
						&wrk_path, &wrk_type);
		} else {
			wrk_error = dist_from_new_tile(nxt_t_ind, slt_tiles,
						C_UP, new_atm, direction);
		}
		if (bst_error > wrk_error) {
			bst_error = wrk_error;
			bst_grad_type = grad_type;
			if (my_settings.ca_only_flag == TRUE) {
				bst_path = wrk_path;
				bst_type = wrk_type;
			} else {
				bst_code = C_UP;
				bst_type = cur_type;
			}
		}
		
		/* Shift the tile up: [pqr] -> [tpq] */
		cur_type = ston[INDX(qtos[cur_type].s, qtos[cur_type].p,
				     qtos[cur_type].q, qtos[cur_type].r)];
	}
	
	if (my_settings.ca_only_flag == TRUE) {
		max_path = my_settings.max_path;
		if ((direction == DIR_BACKWARD && bst_path < max_path / 2)
		    || (direction == DIR_FORWARD && bst_path >= max_path / 2))
			bst_code = C_UP;
		else
			bst_code = C_DOWN;
	}
	slt_tiles[nxt_t_ind].type = cur_type_saved;
	*pbst_code = bst_code;
	*pbst_type = bst_type;
	return bst_grad_type;
}

#define ROTATE_BACKWARD(tgl_ind) (((tgl_ind) <= 0) ? 5 : (tgl_ind) - 1)
#define ROTATE_FORWARD(tgl_ind) (((tgl_ind) >= 5) ? 0 : (tgl_ind) + 1)

/* target atom:  my_protein.atom[atm_ct] */
/* current tile:  wrk_tiles[tgl_ind].type */
/* next tile:     wrk_tiles[1-tgl_ind] */
/* If (direction==Backward)                           */
/* next target_atom:  my_protein.atom[atm_ct - 1] */
/* If (direction==Forward)                             */
/* next target_atom:  my_protein.atom[atm_ct + 1] */
/* Renew the target atom in the case of over-run */
/* nxt_error: distance between the next tile and the target atom */
/* wrk_tiles: current tile, wrk_tiles[6] */
/* atm_ct: target atom */
float seek_next_tile(float nxt_error, int tgl_ind,
		     struct pe_slt_tile * wrk_tiles, int atm_ct,
		     PE_DIRECTION direction)
{
	float nxt_U_error, nxt_D_error;
	float nxtnxt_U_error, nxtnxt_D_error;
	float new_nxt_error;
	int nxt_ind;
	PE_CODE nxt_code;
	
	if (direction == DIR_BACKWARD) {
		nxt_ind = ROTATE_BACKWARD(tgl_ind);
		wrk_tiles[nxt_ind].error = nxt_error;
		wrk_tiles[nxt_ind].error_atom = atm_ct;
		wrk_tiles[nxt_ind].atom_flag = 0;
		wrk_tiles[nxt_ind].dx = my_protein.atom[atm_ct].x - my_protein.atom[atm_ct + 1].x;
		wrk_tiles[nxt_ind].dy = my_protein.atom[atm_ct].y - my_protein.atom[atm_ct + 1].y;
		wrk_tiles[nxt_ind].dz = my_protein.atom[atm_ct].z - my_protein.atom[atm_ct + 1].z;
		set_cen_of_new_tile_back(nxt_ind, wrk_tiles, tgl_ind);
		
		dist_from_next_tile_back(tgl_ind, wrk_tiles,
					 atm_ct, &nxt_U_error, &nxt_D_error);
		if (ALREADY_REACHED(nxt_error) && (atm_ct > 1)
		    && (nxt_U_error > nxt_error) && (nxt_D_error > nxt_error)) {
			dist_from_next_tile_back(tgl_ind, wrk_tiles, atm_ct - 1,
					&nxtnxt_U_error, &nxtnxt_D_error);
			if (nxtnxt_U_error > nxtnxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_UP;	/* Backward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_DOWN;	/* Backward UP */
			}
		} else {
			if (nxt_U_error > nxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_UP;	/* Backward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_DOWN;	/* Backward UP */
			}
		}
		set_top_of_new_tile_back(nxt_ind, wrk_tiles, nxt_code, tgl_ind);
	} else {	/* if (direction == DIR_BACKWARD) { */
		nxt_ind = ROTATE_FORWARD(tgl_ind);
		wrk_tiles[nxt_ind].error = nxt_error;
		wrk_tiles[nxt_ind].error_atom = atm_ct;
		wrk_tiles[nxt_ind].atom_flag = 0;
		wrk_tiles[nxt_ind].dx = my_protein.atom[atm_ct].x - my_protein.atom[atm_ct - 1].x;
		wrk_tiles[nxt_ind].dy = my_protein.atom[atm_ct].y - my_protein.atom[atm_ct - 1].y;
		wrk_tiles[nxt_ind].dz = my_protein.atom[atm_ct].z - my_protein.atom[atm_ct - 1].z;
		set_cen_of_new_tile_for(nxt_ind, wrk_tiles, tgl_ind);
		
		dist_from_next_tile_for(tgl_ind, wrk_tiles,
					atm_ct, &nxt_U_error, &nxt_D_error);
		if (ALREADY_REACHED(nxt_error) && (atm_ct < my_protein.atm_num)
		    && (nxt_U_error > nxt_error) && (nxt_D_error > nxt_error)) {
			dist_from_next_tile_for(tgl_ind, wrk_tiles, atm_ct + 1,
					&nxtnxt_U_error, &nxtnxt_D_error);
			if (nxtnxt_U_error > nxtnxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_DOWN;	/* Forward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_UP;	/* Forward UP */
			}
		} else {
			if (nxt_U_error > nxt_D_error) {
				new_nxt_error = nxt_D_error;
				nxt_code = C_DOWN;	/* Forward DOWN */
			} else {
				new_nxt_error = nxt_U_error;
				nxt_code = C_UP;	/* Forward UP */
			}
		}
		set_top_of_new_tile_for(nxt_ind, wrk_tiles, nxt_code, tgl_ind);
	}	/* if (direction == DIR_BACKWARD) { */
	return new_nxt_error; /* squared distance */
}

/* Choose the best gradient for the initial tile */
/* There are 4 types of gradient: [pqr], [tpq], [stp], qnd [rst]. */
void set_best_init_tile(int nxt_ind, struct pe_slt_tile * slt_tiles,
			int atm_ct, PE_DIRECTION direction, float cur_error)
{
	int bst_type, bst_grad_type;
	PE_CODE bst_code;
	int cur_type;
	int grad_type;
	
	bst_grad_type = best_init_tile_and_code(nxt_ind, slt_tiles, atm_ct,
						direction, &bst_code,
						&bst_type);
	/* Shift the tile upto the best gradient: [pqr] -> [tpq] -> ... */
	cur_type = slt_tiles[nxt_ind].type;
	for (grad_type = 0; grad_type < bst_grad_type; grad_type++) {
		slt_tiles[nxt_ind].top[qtos[cur_type].s]
		= slt_tiles[nxt_ind].top[qtos[cur_type].s] - 1;
		cur_type = ston[INDX(qtos[cur_type].s, qtos[cur_type].p,
				     qtos[cur_type].q, qtos[cur_type].r)];
	}
	slt_tiles[nxt_ind].code = bst_code;
	slt_tiles[nxt_ind].type = bst_type;
	slt_tiles[nxt_ind].tx = slt_tiles[nxt_ind].cx
		- my_trn_table[bst_type].cx;
	slt_tiles[nxt_ind].ty = slt_tiles[nxt_ind].cy
		- my_trn_table[bst_type].cy;
	slt_tiles[nxt_ind].tz = slt_tiles[nxt_ind].cz
		- my_trn_table[bst_type].cz;
	slt_tiles[nxt_ind].error = cur_error;
	slt_tiles[nxt_ind].error_atom = atm_ct;
	slt_tiles[nxt_ind].atom_flag = 1;
	
	slt_tiles[nxt_ind].bx = slt_tiles[nxt_ind].tx
		+ my_trn_table[bst_type].bx;
	slt_tiles[nxt_ind].by = slt_tiles[nxt_ind].ty
		+ my_trn_table[bst_type].by;
	slt_tiles[nxt_ind].bz = slt_tiles[nxt_ind].tz
		+ my_trn_table[bst_type].bz;
	
	if (direction == DIR_BACKWARD) {
		int prv_atm;
		
		prv_atm = seek_previous_target_atom(atm_ct);
		if (prv_atm < 0)
			prv_atm = atm_ct;
		slt_tiles[nxt_ind].dx = my_protein.atom[atm_ct].x
			- my_protein.atom[prv_atm].x;
		slt_tiles[nxt_ind].dy = my_protein.atom[atm_ct].y
			- my_protein.atom[prv_atm].y;
		slt_tiles[nxt_ind].dz = my_protein.atom[atm_ct].z
			- my_protein.atom[prv_atm].z;
	} else {
		int nxt_atm;
		
		nxt_atm = seek_next_target_atom(atm_ct);
		if (nxt_atm < 0)
			nxt_atm = atm_ct;
		slt_tiles[nxt_ind].dx = my_protein.atom[nxt_atm].x
			- my_protein.atom[atm_ct].x;
		slt_tiles[nxt_ind].dy = my_protein.atom[nxt_atm].y
			- my_protein.atom[atm_ct].y;
		slt_tiles[nxt_ind].dz = my_protein.atom[nxt_atm].z
			- my_protein.atom[atm_ct].z;
	}
	if (my_settings.move_to_target == TRUE) {
		move_tile_to_target(nxt_ind, slt_tiles, atm_ct);
	}
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(nxt_ind, slt_tiles);
		if (0) {
			float wrk_x, wrk_y, wrk_z;
			float wrk_length;
			
			wrk_x = (slt_tiles[nxt_ind].cx - slt_tiles[nxt_ind].tx) - my_trn_table[cur_type].cx;
			wrk_y = (slt_tiles[nxt_ind].cy - slt_tiles[nxt_ind].ty) - my_trn_table[cur_type].cy;
			wrk_z = (slt_tiles[nxt_ind].cz - slt_tiles[nxt_ind].tz) - my_trn_table[cur_type].cz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_best_init_tile() : ERROR |slt_tiles[%d].(c* - t*) - my_trn_table[%d].c*| = %8.3f\n",
					nxt_ind, cur_type, wrk_length);
				if (0) print_trn_table();
				exit(1);
			}
			wrk_x = (slt_tiles[nxt_ind].bx - slt_tiles[nxt_ind].tx) - my_trn_table[cur_type].bx;
			wrk_y = (slt_tiles[nxt_ind].by - slt_tiles[nxt_ind].ty) - my_trn_table[cur_type].by;
			wrk_z = (slt_tiles[nxt_ind].bz - slt_tiles[nxt_ind].tz) - my_trn_table[cur_type].bz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_best_init_tile() : ERROR |slt_tiles[%d].(b* - t*) - my_trn_table[%d].b*| = %8.3f\n",
					nxt_ind, cur_type, wrk_length);
				if (0) print_trn_table();
				exit(1);
			}
		}
		
	}
	if (0) {
		fprintf(STDERR,
			"set_best_init_tile() : slt_tiles[%d].error= %f against = %d-th atom\n",
			nxt_ind, slt_tiles[nxt_ind].error,
			slt_tiles[nxt_ind].error_atom);
	}
	return;
}

int set_initial_tile(struct pe_slt_tile * slt_tiles, int ind, int tile_type)
{
	float error_forward_up, error_forward_down;
	float error_backward_up, error_backward_down;
	float bst_error;
	int path_forward_up, path_forward_down;
	int type_forward_up, type_forward_down;
	int path_backward_up, path_backward_down;
	int type_backward_up, type_backward_down;
	int prv_ind, nxt_ind;
	
	if ((ind >= my_protein.atm_num) || (ind <= 1)) {
		fprintf(STDOUT,
			"set_initial_tile(): %d-th atom dosen't exist\n",
			ind);
		return FAIL;
	}
	set_tile_to(slt_tiles, ind, tile_type);
	
	prv_ind = seek_previous_target_atom(ind);
	if (prv_ind < 0) {
		fprintf(STDERR,
			"set_initial_tile() : ERROR There exists no CA atom (from %d to %d).\n",
			prv_ind + 1, ind - 1);
		return FAIL;
	}
	nxt_ind = seek_next_target_atom(ind);
	if (nxt_ind < 0) {
		fprintf(STDERR,
			"set_initial_tile() : ERROR There exists no CA atom (from %d to %d).\n",
			ind - 1, nxt_ind + 1);
		return FAIL;
	}
	
	if (my_settings.ca_only_flag == TRUE) {
		error_backward_down = better_dist_from_npath(0, slt_tiles,
						C_DOWN, prv_ind, DIR_BACKWARD,
						&path_backward_down,
						&type_backward_down);
		error_backward_up = better_dist_from_npath(0, slt_tiles,
						C_UP, prv_ind, DIR_BACKWARD,
						&path_backward_up,
						&type_backward_up);
		if (error_backward_down > error_backward_up) {
			slt_tiles->code = C_UP;
		} else {
			slt_tiles->code = C_DOWN;
		}
		error_forward_down = better_dist_from_npath(0, slt_tiles,
						C_DOWN, nxt_ind, DIR_FORWARD,
						&path_forward_down,
						&type_forward_down);
		error_forward_up = better_dist_from_npath(0, slt_tiles,
						C_UP, nxt_ind, DIR_FORWARD,
						&path_forward_up,
							  &type_forward_up);
		
		bst_error = npath_inconsistency_check(slt_tiles, ind,
					path_forward_down, path_forward_up,
					error_backward_down, error_forward_down,
					error_backward_up, error_forward_up);
	} else {
		error_backward_down = dist_from_new_tile(0, slt_tiles, C_DOWN,
							 prv_ind, DIR_BACKWARD);
		error_backward_up = dist_from_new_tile(0, slt_tiles, C_UP,
						       prv_ind, DIR_BACKWARD);
		if (error_backward_down > error_backward_up) {
			slt_tiles->code = C_UP;
		} else {
			slt_tiles->code = C_DOWN;
		}
		error_forward_down = dist_from_new_tile(0, slt_tiles, C_DOWN,
							nxt_ind, DIR_FORWARD);
		error_forward_up = dist_from_new_tile(0, slt_tiles, C_UP,
						      nxt_ind, DIR_FORWARD);
		
		bst_error = tile_inconsistency_check(slt_tiles, ind,
				error_backward_up, error_forward_up,
				error_backward_down, error_forward_down);
	}
	slt_tiles->bx = slt_tiles->tx
		+ my_trn_table[slt_tiles->type].bx;
	slt_tiles->by = slt_tiles->ty
		+ my_trn_table[slt_tiles->type].by;
	slt_tiles->bz = slt_tiles->tz
		+ my_trn_table[slt_tiles->type].bz;
	slt_tiles->dx = my_protein.atom[nxt_ind].x
		- my_protein.atom[prv_ind].x;
	slt_tiles->dy = my_protein.atom[nxt_ind].y
		- my_protein.atom[prv_ind].y;
	slt_tiles->dz = my_protein.atom[nxt_ind].z
		- my_protein.atom[prv_ind].z;
	if (my_settings.rotate_tile == TRUE) {
		rotate_tile_and_trn_table(0, slt_tiles);
		if (0) {
			float wrk_x, wrk_y, wrk_z;
			float wrk_length;
			int cur_type;
			
			cur_type = slt_tiles[0].type;
			wrk_x = (slt_tiles[0].cx - slt_tiles[0].tx) - my_trn_table[cur_type].cx;
			wrk_y = (slt_tiles[0].cy - slt_tiles[0].ty) - my_trn_table[cur_type].cy;
			wrk_z = (slt_tiles[0].cz - slt_tiles[0].tz) - my_trn_table[cur_type].cz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_initial_tile() : ERROR |slt_tiles[%d].(c* - t*) - my_trn_table[%d].c*| = %8.3f\n",
					0, cur_type, wrk_length);
				if (0) print_trn_table();
				exit(1);
			}
			wrk_x = (slt_tiles[0].bx - slt_tiles[0].tx) - my_trn_table[cur_type].bx;
			wrk_y = (slt_tiles[0].by - slt_tiles[0].ty) - my_trn_table[cur_type].by;
			wrk_z = (slt_tiles[0].bz - slt_tiles[0].tz) - my_trn_table[cur_type].bz;
			wrk_length = sqrt(wrk_x * wrk_x + wrk_y * wrk_y + wrk_z * wrk_z);
			if (FLOAT_NEQ(wrk_length, 1.0)) {
				fprintf(STDERR,
					"set_initial_tile() : ERROR |slt_tiles[%d].(b* - t*) - my_trn_table[%d].b*| = %8.3f\n",
					0, cur_type, wrk_length);
				if (0) print_trn_table();
				exit(1);
			}
		}
	}
	if (0) print_set_initial_tile( ind,
				       error_forward_up, error_forward_down,
				       error_backward_up, error_backward_down,
				       path_forward_up, path_forward_down,
				       path_backward_up, path_backward_down,
				       prv_ind, nxt_ind);
	return SUCCESS;
}

/* target atom:  my_protein.atom[atm_ct]   */
/* current tile:  slt_tiles[t_ind]   */
/* loop6_tiles[] contains the last 5 and current tiles. */
/* It returns the brand-new current tile. */
int seek_init_tile(int t_ind, struct pe_slt_tile * slt_tiles, int atm_ct,
		   PE_DIRECTION direction)
{
	float cur_error, next_error;
	int while_ct;
	struct pe_slt_tile loop6_tiles[6];
	float loop6_error[6];
	int last6_ind;
	int toggle6_ind;	
	int nxt_ind;
	
	cur_error = dist_from_cur_tile(t_ind, slt_tiles, atm_ct);
	cur_error = sqrt(cur_error);
	next_error = dist_from_new_tile(t_ind, slt_tiles, slt_tiles[t_ind].code,
					atm_ct, direction);
	next_error = sqrt(next_error);
	setup_loop6_table(t_ind, slt_tiles, loop6_tiles, loop6_error,
			  cur_error);	/* loop6_error[0] */
	if (0) {
		fprintf(STDOUT,
			"seek_init_tile() : DEBUG initial loop6_tiles[0]\n");
		print_tile(0, loop6_tiles);
	}
	toggle6_ind = 0;	/* 0-1-2-3-4-5-0-1-2- ... */
	last6_ind = 0;
	while_ct = 0;
	while ((NOT_REACHED_YET(cur_error) && cur_error > next_error)
	       || NOT_OVER_RUN(cur_error - next_error)
	       || FAR_FROM_TARGET(cur_error)) {
		cur_error = next_error;
		next_error = seek_next_tile(next_error, toggle6_ind,
					    loop6_tiles, atm_ct, direction);
		next_error = sqrt(next_error);
		toggle6_ind = compute_new_toggle6_index(toggle6_ind, direction);
		loop6_error[toggle6_ind] = cur_error;	/* Add new tile */
		if (while_ct >= 4) {
			if (!FLOAT_NEQ(next_error - loop6_error[last6_ind],
				       my_coor.ave_dis)) {
				if (0) fprintf(STDOUT,
					"seek_init_tile() : DEBUG cur_error= %f, next_error= %f <==> loop6_error[%d]=%f  (while_ct= %d)\n",
					cur_error, next_error, last6_ind,
					loop6_error[last6_ind], while_ct);
				toggle6_ind = choose_best_of_loop6_tiles(t_ind,
						last6_ind, atm_ct, loop6_error,
						next_error);
				break;
			}
			last6_ind = compute_new_toggle6_index(last6_ind, 
							      direction);
		}
		while_ct++;
		if (seek_init_tile_overflow_check(while_ct, last6_ind,
			loop6_error, cur_error, next_error) == FAIL) {
			fprintf(STDERR,
				"seek_init_tile() : DEBUG current loop6_tiles[%d]\n",
				toggle6_ind);
			print_tile(toggle6_ind, loop6_tiles);
			if (direction == DIR_BACKWARD) {
				fprintf(STDERR,
					"seek_init_tile() : DEBUG from my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
					atm_ct + 1, my_protein.atom[atm_ct+1].x, my_protein.atom[atm_ct+1].y, my_protein.atom[atm_ct+1].z);
				fprintf(STDERR,
					"seek_init_tile() : DEBUG   to my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
					atm_ct, my_protein.atom[atm_ct].x, my_protein.atom[atm_ct].y, my_protein.atom[atm_ct].z);
			} else {
				fprintf(STDERR,
					"seek_init_tile() : DEBUG from my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
					atm_ct, my_protein.atom[atm_ct].x, my_protein.atom[atm_ct].y, my_protein.atom[atm_ct].z);
				fprintf(STDERR,
					"seek_init_tile() : DEBUG   to my_protein.atom[%d].* = (%8.3f, %8.3f, %8.3f)\n",
					atm_ct + 1, my_protein.atom[atm_ct+1].x, my_protein.atom[atm_ct+1].y, my_protein.atom[atm_ct+1].z);
			}
			exit(1);
		}
		if (my_settings.move_to_target == TRUE
		    || my_settings.rotate_tile == TRUE)
			break;
	}
	if (0) {
		fprintf(STDOUT,
			"seek_init_tile() : DEBUG cur_error = %8.3f, next_error = %8.3f, while_ct = %d\n",
			cur_error, next_error, while_ct);
		if (direction == DIR_BACKWARD)
			fprintf(STDOUT,
				"seek_init_tile() : DEBUG target atom = %d, current tile = %d  [backward]\n",
				atm_ct, t_ind);
		else
			fprintf(STDOUT,
				"seek_init_tile() : DEBUG target atom = %d, current tile = %d  [forward]\n",
				atm_ct, t_ind);
	}
	if (0) {
		fprintf(STDOUT,
			"seek_init_tile() : DEBUG final loop6_tiles[%d]\n", toggle6_ind);
		print_tile(toggle6_ind, loop6_tiles);
	}
	nxt_ind = copy_init_tile(t_ind, slt_tiles, toggle6_ind, loop6_tiles,
				 direction);
	set_best_init_tile(nxt_ind, slt_tiles, atm_ct, direction, cur_error);
	return nxt_ind;
}
