/*
 *  pe_main.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 6/01/05.
 *  Copyright (C) 2005, 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 *  2006-11-01 add option -l (N-tile code)
 *  2006-11-02 remove redundant if-clause
 *  2006-11-03 set default to FOLDING_WITH_ROT_AND_TRANS (fix print)
 *  2006-11-20 version 0.2 release
 *  2006-11-20 version 0.2 release
 *  2007-01-12 add my_settings.try_all_type (remove TRY_ALL_TYPE)
 *             and set try_all_type = FALSE when -t is specified
 *  2007-01-15 replace 0 with FOLDING_UNSPECIFIED
 *  2007-01-15 set my_settings.code_length = 5  (as default value)
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 *  2007-04-16 option "h" is supported in main()
 *  2007-04-16 version 0.4a release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h> /* getopt */
#include "pe_main.h"
#include "pe_pdb_data.h"
#include "pe_encoding.h"
#include "pe_print.h"

extern int read_pdb_file(char * filename);
extern void encode_protein_structure(struct pe_slt_tile * slt_tiles);
extern void print_protein_code(char * filename, struct pe_slt_tile * s_tiles);
extern void print_udcode(char * filename, struct pe_slt_tile * s_tiles);
extern void code_statistics_analyzer(char * filename, struct pe_slt_tile * s_tiles);
extern void clear_slt_tile(struct pe_slt_tile * pt_dst_tile);
extern void n_tile_code_of_protein(struct pe_slt_tile * slt_tiles);
extern void print_ntile_code(struct pe_protein * pt_protein);	/* 2006-11-01 */

struct pe_protein my_protein;
struct pe_settings my_settings;

void setup_folding_only(void)
{	
	my_settings.ca_only_flag = TRUE;
	my_settings.rotate_tile = FALSE;
	my_settings.move_to_target = FALSE;
	if (1) {
		my_settings.variable_path_length = TRUE;
		my_settings.encoding_for_head = 1.2;	/* 1.2 */
		my_settings.encoding_for_tail = 1.9;	/* 2.7 */
		my_settings.encoding_for_step = 0.1;
	} else {
		my_settings.variable_path_length = FALSE;
		my_settings.encoding_for_head = 1.2;
		my_settings.encoding_for_tail = 2.7;
		my_settings.encoding_for_step = 0.2;
	}
	my_settings.along_backbone = TRUE;
	if (my_settings.initial_atom != NOTSPECIFIED)
		return;
	if (1) {
		my_settings.initial_atom = BESTPOINT;
		fprintf(STDOUT,
			"setup_folding_only() : initial atom not specified\n");
	} else {
		my_settings.initial_atom = MIDPOINT;
		/* Remove 2006-11-02 */
		fprintf(STDOUT,
			"setup_folding_only() : initial atom = midpoint\n");
	}
	/*  Consider XYZW and YZWX as the init tile if TRUE, else only XYZW  */
	my_settings.try_all_type = TRUE;	/* 2007-1-12 */
	return;
}

/* ROT & BACKBONE */
void setup_folding_with_rot(void)
{
	my_settings.along_backbone = FALSE;
	my_settings.move_to_target = FALSE;
	my_settings.ca_only_flag = FALSE;
	my_settings.rotate_tile = TRUE;

	my_settings.encoding_for_head = 1.415;
	my_settings.encoding_for_tail = 1.440;
	my_settings.encoding_for_step = 0.005;
	if (my_settings.initial_atom != NOTSPECIFIED)
		return;
	if (1) {
		my_settings.initial_atom = BESTPOINT;
		fprintf(STDOUT,
			"setup_folding_with_rot() : initial atom not specified\n");
	} else {
		my_settings.initial_atom = MIDPOINT;
		/* Remove 2006-11-02 */
		fprintf(STDOUT,
			"setup_folding_with_rot() : initial atom = midpoint\n");
	}
	return;
	/*  Consider XYZW and YZWX as the init tile if TRUE, else only XYZW  */
	my_settings.try_all_type = TRUE;	/* 2007-1-12 */
}

/* ROT & MOV & INIT & CA */
void setup_folding_with_rot_and_trans(void)
{
	my_settings.along_backbone = FALSE;
	my_settings.move_to_target = TRUE;
	my_settings.res_code_length = 1;
	if (1)	my_settings.ca_only_flag = TRUE;
	else	my_settings.ca_only_flag = FALSE;

	if (1)	my_settings.rotate_tile = TRUE;
	else	my_settings.rotate_tile = FALSE;
	my_settings.encoding_for_head = 1.415;
	my_settings.encoding_for_tail = 1.440;
	my_settings.encoding_for_step = 0.005;
	if (my_settings.initial_atom != NOTSPECIFIED)
		return;
	if (0) {
		my_settings.initial_atom = BESTPOINT;
		fprintf(STDOUT,
			"setup_folding_with_rot_and_trans() : initial atom not specified\n");
	} else {
		my_settings.initial_atom = MIDPOINT;
		/* Remove 2006-11-02 */
		fprintf(STDOUT,
			"setup_folding_with_rot_and_trans() : initial atom = midpoint\n");
	}
	/*  The tyoe of the initial tile is always XYZW */
	my_settings.try_all_type = FALSE;	/* 2007-1-12 */
	return;
}

void setup_parameters(int coding_option)
{	
	/* [TODO] Use getopt_long to decode arguments */
	
	my_settings.model_limit = 1.0;
	my_settings.tiles_per_res = 5.0;
	my_settings.res_code_length = 5;
	my_settings.max_path = -1;
	/* Usage: |pos_a - pos_b| < ave_dis * my_settings.dist_factor */
	my_settings.dist_factor = (1.0 / 2.0);
	my_settings.error_tab_size = 15;
	switch (coding_option) {
		case (FOLDING_ONLY):
			if (DEBUG)
				fprintf(STDOUT, "setup_parameters() : DEBUG folding only\n");
			setup_folding_only();
			break;
		case (FOLDING_WITH_ROTATION):	/* ROT & BACKBONE */
			if (DEBUG)
				fprintf(STDOUT, "setup_parameters() : DEBUG folding with rotation\n");
			setup_folding_with_rot();
			break;
		case (FOLDING_WITH_ROT_AND_TRANS):	/* ROT & MOV & INIT & CA */
			if (DEBUG)
				fprintf(STDOUT, "setup_parameters() : DEBUG folding with rotation and translation\n");
			setup_folding_with_rot_and_trans();
			break;
		default:
			if (DEBUG)
				fprintf(STDOUT,
					"setup_parameters() : DEBUG default (folding with rotation and translation)\n");	/* 2006-11-03 */
			setup_folding_with_rot_and_trans();
	}
	my_settings.print_udcode = TRUE;
	my_settings.stats_ON = FALSE;
	my_settings.statsN_ON = FALSE;
	/*  Consider XYZW and YZWX as the init tile if TRUE, else only XYZW  */
	my_settings.try_all_type = FALSE;	/* 2007-1-12 */

	if (0) {
		fprintf(STDOUT, "[NOTE] Only first %d models are considered\n",
			my_settings.model_limit);
		fprintf(STDOUT, "\n");
	}
	return;
}

#define DEVI_FACTOR 5.0	/* to absorb the deviation of #tile/#residue */

struct pe_slt_tile * allocate_slt_tiles(void)
{
	int max_tiles_per_res;
	struct pe_slt_tile * slt_tiles;
	int  i;
	
	max_tiles_per_res = (int)(DEVI_FACTOR * my_settings.tiles_per_res);
	slt_tiles = (struct pe_slt_tile *) malloc( (my_protein.res_num * 
			max_tiles_per_res + 1) * sizeof(struct pe_slt_tile));
	if (slt_tiles == NULL) {
		fprintf(STDERR, 
			"allocate_slt_tiles() : ERROR allocation failed\n");
		return slt_tiles;
	}
	my_protein.tile_max_num = my_protein.res_num * max_tiles_per_res;
	
	for (i = 0; i <= my_protein.tile_max_num; i++)
		clear_slt_tile(slt_tiles + i);
	return slt_tiles;
}

#define CODE_ANALYSIS_SAFE_LIMIT 24

/**
 * main - encodes the structure of a protein into a binary sequence.
 * @argc   :
 * @argv[1]: encoding option (optional)
 * @argv[2]: input pdb file name
 *
 * main() reads a "***.ent" file and writes the protein code in a 
 * "***.upcode" file and compiles the statistics of the code in a 
 * "***.stats" file.
 *
 * See README.txt for its synopsis and options
 */
int main(int argc, char ** argv)
{
	struct pe_slt_tile * slt_tiles;
	int c;
	int coding_option;
	char * filename;

	coding_option = FOLDING_UNSPECIFIED;	/* 2007-1-15 */
	my_settings.initial_atom = NOTSPECIFIED;
	my_settings.code_length = 5;		/* 2007-1-15 */
	while ((c = getopt(argc, argv, "frthi:l:")) != -1 ) {
		switch (c) {
			case 'f':	/* folding only */
				coding_option = FOLDING_ONLY;
				break;
			case 'r':	/* folding with rotation */
				coding_option = FOLDING_WITH_ROTATION;
				break;
			case 't':	/* folding with rotation and translation */
				coding_option = FOLDING_WITH_ROT_AND_TRANS;
				break;
			case 'i':
				my_settings.initial_atom = atoi(optarg);
				break;
			case 'l':
				my_settings.code_length = atoi(optarg);
				break;
			case 'h':	/* help  2007-04-16 */
				fprintf(STDERR, "Synopsis : ProteinEncoder [-f, -r, or -t] [-i index] [-l length] [-h]  filename\n");
				return 0;
			case '?':
			default:
				fprintf(STDERR, "main() : Unrecognised option : -%c\n", optopt);
				exit(1);
		}
	}

	/* The last arg is the filename */
	filename = argv[argc -1];
	
	fprintf(STDOUT, "INPUT FILE : %s\n", filename);
	setup_parameters(coding_option);

	if (read_pdb_file(filename) == FAIL) exit(1);
	if (0) fprintf(STDOUT, "main() :    ... pdb file has been read\n");
	if (0) print_my_sequences();
	if (LOAD_ONLY) exit(0);

	slt_tiles = allocate_slt_tiles();
	if (slt_tiles == NULL) exit(1);
	
	if (my_settings.code_length > 1) {
		if (DEBUG) {
			fprintf(STDOUT, "\n");
			fprintf(STDOUT, "main() :  Calculate %d-tile code\n",
				my_settings.code_length);
		}
		n_tile_code_of_protein(slt_tiles);
		if (0) print_ntile_code(&my_protein);
	} else {
		encode_protein_structure(slt_tiles);
	}

	print_protein_code(filename, slt_tiles);
	if (0) print_my_protein(&my_protein);
	if (0) print_my_slt_tiles(slt_tiles);
	
	if (my_settings.print_udcode == TRUE)
		print_udcode(filename, slt_tiles);

	if (my_settings.stats_ON == TRUE) {	/* 17/July/2006 */
		if (( (int) my_settings.tiles_per_res < CODE_ANALYSIS_SAFE_LIMIT))
			code_statistics_analyzer(filename, slt_tiles);
		else {
			fprintf(STDOUT,
				"ENCODING: No code analysis because of the SAFE_LIMIT over\n");
			fprintf(STDOUT,
				"ENCODING:     my_settings.tiles_per_res= %8.3f >= %d\n",
				my_settings.tiles_per_res, CODE_ANALYSIS_SAFE_LIMIT);
    		}
	}
	if (slt_tiles != NULL) free(slt_tiles);
	if (my_protein.model != NULL) free(my_protein.model);
	if (my_protein.atom != NULL) free(my_protein.atom);
	if (my_protein.sequence != NULL) free(my_protein.sequence);
	if (my_protein.ntile_code != NULL) free(my_protein.ntile_code);	/* 2006-11-01 */
	exit(0);
}
