/*
 *  pe_n_tile_code_of_protein.c
 *  ProteinEncoder
 *
 *  Created by Naoto Morikawa on 11/1/06.
 *  Copyright (C) 2006, 2007 Naoto Morikawa. All rights reserved.
 *
 *  2006-11-20 version 0.2 release
 *  2007-02-15 version 0.3 release
 *  2007-04-15 version 0.4 release
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "pe_main.h"
#include "pe_pdb_data.h"	/* struct pe_counter */
#include "pe_encoding.h"
#include "pe_print.h"

extern void initialize_protein_data_arrays(void);
extern int allocate_protein_data_arrays(struct pe_counter * p_ct);
extern void encode_protein_structure(struct pe_slt_tile * slt_tiles);

extern struct pe_protein my_protein;
extern struct pe_settings my_settings;

void print_ntile_code(struct pe_protein * pt_protein)
{
	int mod_ct, seq_ct, atm_ct, t_ct;
	
	fprintf(STDOUT, "\n");
	fprintf(STDOUT,
		"<N-tile_code atoms=\"%d\" residues=\"%d\" sequences=\"%d\" molecules=\"%d\" models=\"%d\">\n",
		pt_protein->atm_num, pt_protein->res_num, pt_protein->seq_num,
		pt_protein->mol_num, pt_protein->mod_num);
	for (mod_ct = 1; mod_ct <= pt_protein->mod_num; mod_ct++) {
		for (seq_ct = pt_protein->model[mod_ct].seq_head;
		     seq_ct <= pt_protein->model[mod_ct].seq_tail; seq_ct++) {
			fprintf(STDOUT,
				"<N-tile_sequence No.=\"%d\" from=\"%d-th atom\" to=\"%d-th atom\">\n",
				seq_ct, pt_protein->sequence[seq_ct].atom_head,
				pt_protein->sequence[seq_ct].atom_tail);
			fprintf(STDOUT,
				"        [  No. ]  ser.  name res chn num  N-tile_code\n");
			for (atm_ct = pt_protein->sequence[seq_ct].atom_head;
			     atm_ct <= pt_protein->sequence[seq_ct].atom_tail;
			     atm_ct++) {
				fprintf(STDOUT, "<ntile> [%6d] %6d ", atm_ct,
					pt_protein->atom[atm_ct].serial_num);
				if (pt_protein->atom[atm_ct].atom_name == PE_CC)
					fprintf(STDOUT, " C  ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_CA)
					fprintf(STDOUT, " CA ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_NN)
					fprintf(STDOUT, " N  ");
				else if (pt_protein->atom[atm_ct].atom_name
					 == PE_RNA)	/* DNA or RNA */
					 fprintf(STDOUT, "    ");
				else
					 fprintf(STDOUT, " UN ");
				if (pt_protein->atom[atm_ct].chain_id == ' ') {
					/* To use sscanf() */
					pt_protein->atom[atm_ct].chain_id = '-';
				}
				fprintf(STDOUT, " %.3s  %c  %4d ",
					pt_protein->atom[atm_ct].res_name,
					pt_protein->atom[atm_ct].chain_id,
					pt_protein->atom[atm_ct].res_seq);
				t_ct = pt_protein->atom[atm_ct].ntile_head;
				if (t_ct >= 0) {
					for (; t_ct < pt_protein->atom[atm_ct].ntile_head
					     + pt_protein->atom[atm_ct].ntile_length;
					     t_ct++) {
						if (pt_protein->ntile_code[t_ct] == C_UP)
							fprintf(STDOUT, "U");
						else if (pt_protein->ntile_code[t_ct] == C_DOWN)
							fprintf(STDOUT, "D");
						else
							fprintf(STDOUT, "X");
					}
				} else {
					for (; t_ct < pt_protein->atom[atm_ct].ntile_head
					     + my_settings.code_length;
					     t_ct++) {
						fprintf(STDOUT, "-");
					}
				}
				
				fprintf(STDOUT, " </ntile>\n");
			}
			fprintf(STDOUT, "</N-tile_sequence>\n");
			if (seq_ct < pt_protein->model[mod_ct].seq_tail)
				fprintf(STDOUT, "\n");
		}
		break;
	}
	return;
}

int search_sequence2(int ind /* atom No. */, struct pe_protein * pt_protein)
{
	int mod_ct, seq_ct;
	
	for (mod_ct = 1; mod_ct <= pt_protein->mod_num; mod_ct++) {
		for (seq_ct = pt_protein->model[mod_ct].seq_head;
		     seq_ct <= pt_protein->model[mod_ct].seq_tail; seq_ct++) {
			if ((pt_protein->sequence[seq_ct].atom_head <= ind)
			    && (ind <= pt_protein->sequence[seq_ct].atom_tail)) {
				return seq_ct;
			}
		}
	}
	fprintf(STDERR, "search_sequence2() : The %d-th atom does not exists.\n",
		ind);
	return -1;
}

void mv_protein(struct pe_protein * pt_dst, struct pe_protein * pt_src)
{
	int i;
	
	pt_dst->mod_num = pt_src->mod_num;
	pt_dst->model = pt_src->model;

	pt_dst->mol_num = pt_src->mol_num;
	for (i = 1; i <= pt_src->mol_num; i++) {
		pt_dst->molecule[i].chain_head = pt_src->molecule[i].chain_head;
		pt_dst->molecule[i].chain_tail = pt_src->molecule[i].chain_tail;
		pt_dst->molecule[i].id = pt_src->molecule[i].id;
		strncpy(pt_dst->molecule[i].name, pt_src->molecule[i].name, LINE_WIDTH + 1);
	}
	
	pt_dst->chn_num = pt_src->chn_num;
	for (i = 1; i <= pt_src->chn_num; i++) {
		pt_dst->chain[i].atom_head = pt_src->chain[i].atom_head;
		pt_dst->chain[i].atom_tail = pt_src->chain[i].atom_tail;
		pt_dst->chain[i].id = pt_src->chain[i].id;
		pt_dst->chain[i].mol_id = pt_src->chain[i].mol_id;
	}
	
	pt_dst->res_num = pt_src->res_num;
	pt_dst->atm_num = pt_src->atm_num;
	pt_dst->atom = pt_src->atom;
	
	pt_dst->seq_num = pt_src->seq_num;
	pt_dst->sequence = pt_src->sequence;
	pt_dst->tile_max_num = pt_src->tile_max_num;
	pt_dst->tile_num = pt_src->tile_num;
	
	pt_dst->best_ind = pt_src->best_ind;
	pt_dst->best_t_type = pt_src->best_t_type;
	pt_dst->ave_bond_length = pt_src->ave_bond_length;
	pt_dst->ave_dis = pt_src->ave_dis;
	pt_dst->ave_error = pt_src->ave_error;
	
	pt_dst->ntile_code = pt_src->ntile_code;
	pt_dst->ntile_top = pt_src->ntile_top;
	pt_dst->ntile_size = pt_src->ntile_size;
	
	pt_src->mod_num = 0;
	pt_src->model = NULL;
	pt_src->mol_num = 0;
	pt_src->chn_num = 0;
	pt_src->res_num = 0;
	pt_src->atm_num = 0;
	pt_src->atom = NULL;
	pt_src->seq_num = 0;
	pt_src->sequence = NULL;
	pt_src->tile_num = 0;
	pt_src->ntile_code = NULL;
	pt_src->ntile_top = -1;
	pt_src->ntile_size = 0;
	
	pt_src->best_ind = -1;
	pt_src->best_t_type = -1;
	pt_src->ave_bond_length = -1.0;
	pt_src->ave_dis = -1.0;
	pt_src->ave_error = -1.0;
	return;
}

void setup_dummy_count(struct pe_counter * pt_count)
{
	pt_count->model = 1;
	pt_count->molecule = 1;
	pt_count->chain = 1;
	pt_count->residue = my_settings.code_length;	/* Count of CA atoms */
	pt_count->sequence = 1;	/* count of TER */
	pt_count->atom = 0;
	return;
}

void set_my_protein_to_n_fragment(void)
{	
	my_protein.atm_num = 0;
	my_protein.molecule[1].chain_head = 1;
	my_protein.molecule[1].chain_tail = 1;
	my_protein.molecule[1].id = 1;
	strncpy(my_protein.molecule[1].name, "N-fragment\0", LINE_WIDTH + 1);
	
	my_protein.chain[1].atom_head = 0;
	my_protein.chain[1].atom_tail = 0;
	my_protein.chain[1].tile_head = 0;
	my_protein.chain[1].tile_tail = 0;
	my_protein.chain[1].id = '-';
	my_protein.chain[1].mol_id = 1;
	my_protein.chain[1].mod_num = 1;
	
	my_protein.sequence[1].mol_id = 1;
	my_protein.sequence[1].mod_num = 1;
	my_protein.model[1].seq_head = 1;
	my_protein.model[1].seq_tail = 1;
	
	return;
}

int allocate_ntile_code(struct pe_protein * pt_protein)
{
	PE_CODE * ntile_code;
	int array_size;
	int mod_ct, seq_ct, atm_ct, i;
	
	array_size = pt_protein->atm_num * my_settings.code_length
		* (my_settings.tiles_per_res + 3);
	ntile_code = (PE_CODE *) malloc(array_size * sizeof(PE_CODE));
	if (ntile_code == NULL) {
		fprintf(STDERR, 
			"allocate_ntile_code() : ERROR allocation failed\n");
		return FAIL;
	}
	for (i = 0; i < array_size; i++) {
		ntile_code[i] = C_UNKNOWN;
	}
	pt_protein->ntile_code = ntile_code;
	pt_protein->ntile_top = 0;
	pt_protein->ntile_size = array_size;
	
	for (mod_ct = 1; mod_ct <= pt_protein->mod_num; mod_ct++) {
		for (seq_ct = pt_protein->model[mod_ct].seq_head;
		     seq_ct <= pt_protein->model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = pt_protein->sequence[seq_ct].atom_head;
			     atm_ct <= pt_protein->sequence[seq_ct].atom_tail;
			     atm_ct++) {
				pt_protein->atom[atm_ct].ntile_head = -1;
				pt_protein->atom[atm_ct].ntile_length = 0;
			}
		}
	}
	return SUCCESS;
}

#define THIS_IS_NOT_CA_ATOM_2(atm_ct, wrk_protein) \
(my_settings.ca_only_flag == TRUE \
	&& wrk_protein.atom[(atm_ct)].atom_name != PE_CA)

/* cp_fragment_and_set_init_atom() specifies the value of my_settings.initial_atom. */
int cp_fragment_and_set_init_atom(struct pe_protein * pt_dst, struct pe_protein * pt_src, int atm_ind)
{
	int half_length;
	int atm_ct, a_ct;
	int seq_ct;
	int tile_ct;
	int top_atm, end_atm;
	
	half_length = (my_settings.code_length - 1) / 2;
	seq_ct = search_sequence2(atm_ind, pt_src);
	top_atm = -1;
	tile_ct = 0;
	for (atm_ct = atm_ind - 1;
	     atm_ct >= pt_src->sequence[seq_ct].atom_head;
	     atm_ct--) {
		if (THIS_IS_NOT_CA_ATOM_2(atm_ct, (*pt_src))) continue;
		tile_ct++;
		if (tile_ct >= half_length) {
			top_atm = atm_ct;
			break;
		}
		
	}
	if (top_atm < 0) return FAIL;
	end_atm = -1;
	tile_ct = 0;
	for (atm_ct = atm_ind + 1;
	     atm_ct <= pt_src->sequence[seq_ct].atom_tail;
	     atm_ct++) {
		if (THIS_IS_NOT_CA_ATOM_2(atm_ct, (*pt_src))) continue;
		tile_ct++;
		if (tile_ct >= half_length) {
			end_atm = atm_ct;
			break;
		}
	}
	if (end_atm < 0) return FAIL;
	
	if (0) {
		fprintf(STDOUT,
			"cp_fragment_and_set_init_atom() : DEBUG [top_atom, atm_ind, end_atm] = [%d, %d, %d]\n",
			top_atm, atm_ind, end_atm);
	}
	a_ct = 1;
	for (atm_ct = top_atm; atm_ct <= end_atm; atm_ct++, a_ct++) {
		if (pt_src->atom[atm_ct].atom_name == PE_CC
		    || pt_src->atom[atm_ct].atom_name == PE_CA
		    || pt_src->atom[atm_ct].atom_name == PE_NN) {
			pt_dst->atom[a_ct].serial_num = pt_src->atom[atm_ct].serial_num;
			pt_dst->atom[a_ct].atom_name = pt_src->atom[atm_ct].atom_name;
			pt_dst->atom[a_ct].alt_loc = pt_src->atom[atm_ct].alt_loc;
			strncpy(pt_dst->atom[a_ct].res_name, pt_src->atom[atm_ct].res_name, 3);
			pt_dst->atom[a_ct].chain_id = pt_src->atom[atm_ct].chain_id;
			pt_dst->atom[a_ct].res_seq = pt_src->atom[atm_ct].res_seq;
			pt_dst->atom[a_ct].x = pt_src->atom[atm_ct].x;
			pt_dst->atom[a_ct].y = pt_src->atom[atm_ct].y;
			pt_dst->atom[a_ct].z = pt_src->atom[atm_ct].z;
			pt_dst->atom[a_ct].occupancy = pt_src->atom[atm_ct].occupancy;
			strncpy(pt_dst->atom[a_ct].seg_id, pt_src->atom[atm_ct].seg_id, 4);
		}
		if (atm_ct == atm_ind) my_settings.initial_atom = a_ct;
	}

	pt_dst->atm_num = a_ct - 1;
	pt_dst->chain[1].atom_head = 1;
	pt_dst->chain[1].atom_tail = pt_dst->atm_num;
	pt_dst->sequence[1].atom_head = 1;
	pt_dst->sequence[1].atom_tail = pt_dst->atm_num;
	return SUCCESS;
}

void set_ntile_code(struct pe_protein * pt_protein, struct pe_slt_tile * slt_tiles, int atm_ct)
{
	int i;
	
	pt_protein->atom[atm_ct].ntile_head = pt_protein->ntile_top;
	for (i = 1; i <= my_protein.tile_num; i++) {
		if (pt_protein->ntile_top >= pt_protein->ntile_size) {
			fprintf(STDERR,
				"set_ntile_code() : ERROR ntile overflow (top= %d >=  size= %d\n",
				pt_protein->ntile_top, pt_protein->ntile_size);
			exit(1);
		}
		pt_protein->ntile_code[pt_protein->ntile_top] = slt_tiles[i].code;
		pt_protein->ntile_top++;
		pt_protein->atom[atm_ct].ntile_length++;
	}
	return;
}

void n_tile_code_of_protein(struct pe_slt_tile * slt_tiles)
{
	int half_length;
	int work;
	int mod_ct, seq_ct, atm_ct;
	struct pe_protein save_my_protein;
	struct pe_counter dummy_count;
	
	/* Check lenghth N whether it is an odd number */
	half_length = (my_settings.code_length - 1) / 2;
	work = half_length * 2 + 1;
	if (work != my_settings.code_length) {
		fprintf(STDOUT,
			"n_tile_code_of_protein() : ERROR n-tile code length = %d is not odd\n",
			my_settings.code_length);
		exit(1);
	}
	
	/* Save my_protein and create a new my_protein of N amino-acids */
	if (0) {
		fprintf(STDOUT,
			"n_tile_code_of_protein() : original my_protein\n");
		print_my_protein(&my_protein);
	}
	mv_protein(&save_my_protein, &my_protein);
	setup_dummy_count(&dummy_count);
	if (allocate_protein_data_arrays(&dummy_count) == FAIL) {
		fprintf(STDERR, "n_tile_code_of_protein() : ERROR n-tile fragment not allocated\n");
		exit(1);
	}
	set_my_protein_to_n_fragment();
	if (allocate_ntile_code(&save_my_protein) == FAIL)
		exit(1);
	
	if (0) {
		fprintf(STDOUT,
			"n_tile_code_of_protein() : new my_protein has been allocated successfully.\n");
		print_my_protein(&my_protein);
	}
	if (0) {
		fprintf(STDOUT,
			"n_tile_code_of_protein() : save_my_protein\n");
		print_my_protein(&save_my_protein);
	}
	
	for (mod_ct = 1; mod_ct <= save_my_protein.mod_num; mod_ct++) {
		for (seq_ct = save_my_protein.model[mod_ct].seq_head;
		     seq_ct <= save_my_protein.model[mod_ct].seq_tail; seq_ct++) {
			for (atm_ct = save_my_protein.sequence[seq_ct].atom_head;
			     atm_ct <= save_my_protein.sequence[seq_ct].atom_tail;
			     atm_ct++) {
				if (THIS_IS_NOT_CA_ATOM_2(atm_ct, save_my_protein))
					continue;
				/* Extract fragments of lenghth N */
				if (cp_fragment_and_set_init_atom(&my_protein,
					&save_my_protein, atm_ct) == SUCCESS) {
					if (0) {
						print_my_protein(&my_protein);
						print_my_sequences();
					}
					encode_protein_structure(slt_tiles);
					if (0) print_my_slt_tiles(slt_tiles);
					/* Set ntile_code of the current atom */
					set_ntile_code(&save_my_protein, slt_tiles, atm_ct);

				} else {
					/* First and last (N-1)/2 atoms */
					if (0) fprintf(STDOUT, "n_tile_code_of_protein() : Skip\n");
				}
			}
		}
		break;	/* Only the first NMR model is analyzed. */
	}
	if (0) print_ntile_code(&save_my_protein);
	
	/* Free n_fragment and restore my_protein */
	if (my_protein.model != NULL) free(my_protein.model);
	if (my_protein.atom != NULL) free(my_protein.atom);
	if (my_protein.sequence != NULL) free(my_protein.sequence);
	if (my_protein.ntile_code != NULL) free(my_protein.ntile_code);
	mv_protein(&my_protein, &save_my_protein);
	
	return;
}
